import os

import boto3

from cloud_governance.common.clouds.aws.utils.utils import Utils


class IAMOperations:

    def __init__(self, iam_client=None):
        self.iam_client = iam_client if iam_client else boto3.client('iam')
        self.utils = Utils()
        self.__sts_client = sts_client = boto3.client('sts')

    def get_user_tags(self, username: str):
        """
        This method return tags from the iam resources
        @param username:
        @return:
        """
        try:
            user = self.iam_client.get_user(UserName=username)['User']
            if user.get('Tags'):
                return user.get('Tags')
            else:
                return []
        except:
            return []

    def get_roles(self):
        """
        This method returns all roles
        @return:
        """
        return self.utils.get_details_resource_list(func_name=self.iam_client.list_roles, input_tag='Roles', check_tag='Marker')

    def get_users(self):
        """
        This method returns all users
        @return:
        """
        return self.utils.get_details_resource_list(self.iam_client.list_users, input_tag='Users', check_tag='Marker')

    def get_account_alias_cloud_name(self):
        """
        This method returns the aws account alias and cloud name
        @return:
        """
        try:
            account_alias = self.iam_client.list_account_aliases()['AccountAliases']
            if account_alias:
                return account_alias[0].upper(), 'AwsCloud'.upper()
        except:
            return os.environ.get('account', '').upper(), 'AwsCloud'.upper()

    def get_iam_users_list(self):
        """
        This method return the IAM users list
        :return:
        """
        iam_users = []
        users = self.get_users()
        for user in users:
            iam_users.append(user.get('UserName'))
        return iam_users

    def get_aws_account_id_name(self):
        """
        This method returns the aws account_id
        :return:
        """
        response = self.__sts_client.get_caller_identity()
        account_id = response['Account']
        return account_id



import datetime

from cloud_governance.policy.policy_operations.aws.zombie_non_cluster.run_zombie_non_cluster_policies import NonClusterZombiePolicy
from cloud_governance.common.clouds.aws.price.price import AWSPrice


class SkippedResources(NonClusterZombiePolicy):
    """
    This class fetches the resources and filter the tags if tag "NOTDELETE" or "SKIP" is present and uploads to the ElasticSearch
    """

    def __init__(self):
        super().__init__()
        self.es_index = 'cloud-governance-skipped-resources'
        self._aws_price = AWSPrice()
        self._ebs_prices = self.get_volume_type_prices()

    def get_volume_type_prices(self):
        """
        This method returns the prices of each ebs volume type
        @return:
        """
        volume_types = ['gp', 'io', 'st1', 'sc1', 'standard']
        volume_types_cost = {}
        for volume_type in volume_types:
            volume_types_cost[volume_type] = self._aws_price.get_ebs_cost(volume_type=volume_type, region=self._region)
        return volume_types_cost

    def get_resources(self, resource_name: str):
        """
        This method returns resource data based on resource name
        @param resource_name:
        @return:
        """
        if resource_name == 'Instance':
            reservations = self._ec2_operations.get_instances()
            return [resource for instances in reservations for resource in instances['Instances']]
        elif resource_name == 'Volume':
            return self._ec2_operations.get_volumes()
        elif resource_name == 'ElasticIp':
            return self._ec2_operations.get_elastic_ips()
        else:
            if resource_name == 'NatGateway':
                return self._ec2_operations.get_nat_gateways()

    def get_ebs_cost(self, volume_id: str):
        """
        This method returns the size of the ebs_volume
        @param volume_id:
        @return:
        """
        volume_types = {'standard': 'standard', 'io1': 'io', 'io2': 'io', 'gp2': 'gp', 'gp3': 'gp', 'sc1': 'sc1',
                        'st1': 'st1'}
        volume = self._ec2_client.describe_volumes(VolumeIds=[volume_id])['Volumes'][0]
        volume_type = volume.get('VolumeType')
        return volume.get('Size') * float(self._ebs_prices[volume_types[volume_type]])

    def get_instance_volume_size(self, resource: dict):
        """
        This method returns size of the attached volumes of the instance
        @param resource:
        @return:
        """
        stopped_time = self._cloudtrail.get_stop_time(resource_id=resource.get('InstanceId'), event_name='StopInstances')
        if stopped_time:
            calculate_days = self._calculate_days(create_date=stopped_time)
        else:
            calculate_days = 90
        block_device = resource.get('BlockDeviceMappings')
        volume_cost = []
        for volume in block_device:
            if volume.get('Ebs'):
                volume_cost.append(self.get_ebs_cost(volume_id=volume.get('Ebs').get('VolumeId')))
        return sum(volume_cost) * (calculate_days/30)

    def get_not_delete_resources(self):
        """
        This method gets all tag "Not_Delete" or "skip" resources
        @return:
        """
        not_delete_resources = []
        resources = {'Instance': 'InstanceId', 'Volume': 'VolumeId', 'ElasticIp': 'AllocationId', 'NatGateway': 'NatGatewayId'}
        for resource_name, resource_id in resources.items():
            delete_resources_data = self.get_resources(resource_name=resource_name)
            for resource in delete_resources_data:
                tags = resource.get('Tags')
                if tags:
                    user = self._get_tag_name_from_tags(tags=tags, tag_name='User')
                    if self._get_policy_value(tags=tags) in ('NOTDELETE', 'SKIP'):
                        resource_data = {'ResourceName': resource_name, 'ResourceId': resource.get(resource_id),
                                         'Region': self._region, 'User': user}
                        volume_cost = 0
                        if resource_name == 'Instance':
                            volume_cost = self.get_instance_volume_size(resource=resource)
                            resource_type = resource.get('InstanceType')
                        else:
                            if resource_name == 'Volume':
                                volume_cost = self.get_ebs_cost(volume_id=resource.get(resource_id))
                                resource_type = resource.get('VolumeType')
                        if volume_cost:
                            resource_data['Cost'] = volume_cost
                        resource_data['ResourceName'] = resource_type
                        not_delete_resources.append(resource_data)

        return not_delete_resources

    def run(self):
        """
        This method returns all tag "Not_Delete" or "skip" resources
        @return:
        """
        resources_data = self.get_not_delete_resources()
        if self._es_upload.es_host:
            end_datetime = datetime.datetime.now().replace(hour=0, minute=0, second=0, microsecond=0)
            start_datetime = end_datetime - datetime.timedelta(1)
            # deleting past data to show fresh report
            self._es_upload.elastic_search_operations.delete_data_in_between_in_es(es_index=self.es_index, start_datetime=start_datetime, end_datetime=end_datetime)
            self._es_upload.es_upload_data(items=resources_data, es_index=self.es_index)
        for resource in resources_data:
            if resource.get('timestamp'):
                resource['timestamp'] = str(resource['timestamp'])
        return resources_data



import datetime


from cloud_governance.common.clouds.aws.cloudwatch.cloudwatch_operations import CloudWatchOperations
from cloud_governance.policy.policy_operations.aws.zombie_non_cluster.run_zombie_non_cluster_policies import NonClusterZombiePolicy


class UnusedNatGateway(NonClusterZombiePolicy):
    """
    This class sends an alert mail for zombie Nat gateways ( based on vpc routes )
    to the user after 4 days and delete after 7 days.
    """

    NAMESPACE = 'AWS/NATGateway'
    UNUSED_DAYS = 1

    def __init__(self):
        super().__init__()
        self._cloudwatch = CloudWatchOperations(region=self._region)

    def __check_cloud_watch_logs(self, resource_id: str, days: int = UNUSED_DAYS):
        """
        This method returns weather the NatGateway is used in last input days
        :param resource_id:
        :param days:
        :return:
        """
        if days == 0:
            days = 1
        end_time = datetime.datetime.utcnow()
        start_time = end_time - datetime.timedelta(days=days)
        response = self._cloudwatch.get_metric_data(start_time=start_time, end_time=end_time, resource_id=resource_id,
                                                    resource_type='NatGatewayId', namespace=self.NAMESPACE,
                                                    metric_names={'ActiveConnectionCount': 'Count'},
                                                    statistic='Average')['MetricDataResults'][0]
        for value in response.get('Values'):
            if value > 0:
                return False
        return True

    def __check_nat_gateway_in_routes(self, nat_gateway_id: str):
        """
        This method check the nat gateway present in the routes or not.
        :param nat_gateway_id:
        :return:
        """
        route_tables = self._ec2_client.describe_route_tables()['RouteTables']
        nat_gateway_found = False
        for route_table in route_tables:
            for route in route_table.get('Routes'):
                if route.get('NatGatewayId') == nat_gateway_id:
                    nat_gateway_found = True
        return nat_gateway_found

    def run(self):
        """
        This method returns zombie NatGateways, delete if dry_run no
        @return:
        """
        nat_gateways = self._ec2_operations.get_nat_gateways()
        nat_gateway_unused_data = []
        for nat_gateway in nat_gateways:
            if self._get_policy_value(tags=nat_gateway.get('Tags', [])) not in ('NOTDELETE', 'SKIP'):
                nat_gateway_id = nat_gateway.get('NatGatewayId')
                tags = nat_gateway.get('Tags')
                gateway_unused = False
                last_used_days = int(self._ec2_operations.get_tag_value_from_tags(tags=tags, tag_name='LastUsedDay', default_value=1))
                if not self._check_cluster_tag(tags=tags):
                    if nat_gateway.get('State') == 'available':
                        if not self.__check_nat_gateway_in_routes(nat_gateway_id=nat_gateway_id) or self.__check_cloud_watch_logs(days=last_used_days, resource_id=nat_gateway_id):
                            gateway_unused = True
                            unused_days = self._get_resource_last_used_days(tags=tags)
                            zombie_nat_gateway = self._check_resource_and_delete(resource_name='NatGateway',
                                                                                 resource_id='NatGatewayId',
                                                                                 resource_type='CreateNatGateway',
                                                                                 resource=nat_gateway,
                                                                                 empty_days=unused_days,
                                                                                 days_to_delete_resource=self.DAYS_TO_DELETE_RESOURCE,
                                                                                 tags=tags)
                            if zombie_nat_gateway:
                                nat_gateway_unused_data.append(
                                    {'ResourceId': nat_gateway_id,
                                     'Name': self._get_tag_name_from_tags(tags=tags, tag_name='Name'),
                                     'User': self._get_tag_name_from_tags(tags=tags, tag_name='User'),
                                     'VpcId': zombie_nat_gateway.get('VpcId'),
                                     'Skip': self._get_policy_value(tags=tags),
                                     'Days': unused_days, 'Policy': self._policy})
                        else:
                            unused_days = 0
                        self._update_resource_tags(resource_id=nat_gateway_id, tags=tags, left_out_days=unused_days,
                                                   resource_left_out=gateway_unused)
        return nat_gateway_unused_data



import datetime
from ast import literal_eval

from cloud_governance.common.clouds.ibm.account.ibm_account import IBMAccount
from cloud_governance.common.logger.init_logger import logger
from cloud_governance.common.mails.mail_message import MailMessage
from cloud_governance.common.mails.postfix import Postfix
from cloud_governance.main.environment_variables import environment_variables


class IBMCostOverUsage:
    """
    This class fetches the ibm cost usage weekly and alert the admins
    """

    EXCLUDE_RESOURCES = ['bare metal servers and attached services', 'virtual server for vpc']
    MAXIMUM_THRESHOLD = 1000

    def __init__(self):
        super().__init__()
        self.__environment_variables_dict = environment_variables.environment_variables_dict
        self.__ibm_account = IBMAccount()
        self.__mail_message = MailMessage()
        self.__mail = Postfix()
        self.__maximum_threshold = self.__environment_variables_dict.get('MAXIMUM_THRESHOLD', '')
        if self.__maximum_threshold:
            self.__maximum_threshold = self.MAXIMUM_THRESHOLD
        self._to_mail = literal_eval(self.__environment_variables_dict.get('to_mail', '[]'))
        self._cc_mail = literal_eval(self.__environment_variables_dict.get('cc_mail', '[]'))

    def get_current_usage(self, month: int, year: int):
        """
        This method gets current usage
        @return:
        """
        usage_data = self.__ibm_account.get_daily_usage(month=month, year=year)
        if usage_data:
            usage_resources = usage_data.get('account_resources')
            usage_details = []
            for resource in usage_resources:
                usage_details.append({'Name': resource.get('resource_name'), 'Cost': resource.get('billable_cost')})
            return usage_details
        return []

    def check_cost_over_usage(self):
        """
        This method check the over usage of the resources
        @return:
        """
        date = datetime.datetime.now()
        month, year = date.month, date.year
        usage_resources = self.get_current_usage(month=month, year=year)
        cost_over_usage_resources = []
        for resource in usage_resources:
            if resource.get('Name').lower() not in self.EXCLUDE_RESOURCES:
                if resource.get('Cost') > int(self.__maximum_threshold):
                    cost_over_usage_resources.append(resource)
        if cost_over_usage_resources:
            subject, body = self.__mail_message.ibm_cost_over_usage(data=self.format_into_html_data(resources=cost_over_usage_resources), month=month, year=year)
            self.__mail.send_email_postfix(subject=subject, to=self._to_mail, cc=self._cc_mail, content=body, message_type='alert_admin', mime_type='html')
            logger.info('Mail sent successfully')

    def format_into_html_data(self, resources: list):
        """
        This method format the data into HTML list form
        @param resources:
        @return:
        """
        start_dt, end_dt = '<dt>', '</dt>'
        start_dl, end_dl = '<dd>', '</dd>'
        start_b, end_b = '<b>', '</b>'
        html_list = ['<dl>']
        for resource in resources:
            start_list = f'{start_dt}{start_b}{resource.get("Name")}{end_b}{end_dt}'
            start_list += f'{start_dl}- {resource.get("Cost")} ${end_dl}'
            html_list.append(start_list)
        html_list.append('</dt>')
        return "\n".join(html_list)

    def run(self):
        """
        This method run the ibm cost_over usage
        @return:
        """
        self.check_cost_over_usage()




from cloud_governance.common.clouds.aws.price.price import AWSPrice
from cloud_governance.main.environment_variables import environment_variables


class ResourcesPricing:
    """
    This class calculates the AWS resources pricing
    """

    MONTHLY_HOURS = 730
    IP_HOURLY_COST = 0.005

    def __init__(self):
        self.__environment_variables_dict = environment_variables.environment_variables_dict
        self._aws_pricing = AWSPrice()
        self.region = self.__environment_variables_dict.get('AWS_DEFAULT_REGION', 'us-east-1')

    def ec2_instance_type_cost(self, instance_type: str, hours: float):
        """
        This method returns the cost of ec2 instance types cost
        @return:
        """
        cost = float(self._aws_pricing.get_price(instance=instance_type, os='Linux', region=self._aws_pricing.get_region_name(self.region)))
        return cost * hours

    def get_ebs_cost(self, volume_size: int, volume_type: str, hours: float):
        """
        This method returns the cost of ebs_volume
        @param hours:
        @param volume_size:
        @param volume_type:
        @return:
        """
        cost = float(self._aws_pricing.get_ebs_cost(volume_type=volume_type, region=self.region))
        return cost * volume_size * (hours / self.MONTHLY_HOURS)

    def get_const_prices(self, resource_type: str, hours: int):
        """
        This method gives the cost of const resources
        @param hours:
        @param resource_type:
        @return:
        """
        if resource_type == 'eip':
            return self.IP_HOURLY_COST * hours




import copy
import datetime
import logging
from ast import literal_eval

import boto3

from cloud_governance.common.clouds.aws.iam.iam_operations import IAMOperations
from cloud_governance.common.clouds.aws.savingsplan.savings_plans_operations import SavingsPlansOperations
from cloud_governance.common.logger.logger_time_stamp import logger_time_stamp

from cloud_governance.common.logger.init_logger import logger, handler
from cloud_governance.common.clouds.aws.cost_explorer.cost_explorer_operations import CostExplorerOperations
from cloud_governance.policy.aws.cost_billing_reports import CostBillingReports


class CostExplorerPayerBillings(CostBillingReports):
    """
    This class is responsible for generation cost billing report for Budget, Actual, Forecast from the Org level
    Monthly savings Plan Amortization: (linked_account_total_cost/payer_account_total_cost) * monthly_savings_plan_cost
    Monthly_support_fee: (monthly_support_fee - (monthly_support_fee * discount ) ) * (linked_account_total_cost/payer_account_total_cost)
    """

    DEFAULT_ROUND_DIGITS = 3

    def __init__(self):
        super().__init__()
        self.__aws_role = self._environment_variables_dict.get("AWS_ACCOUNT_ROLE")
        self.__access_key, self.__secret_key, self.__session = self.__get_sts_credentials()
        self.__ce_client = boto3.client('ce', aws_access_key_id=self.__access_key, aws_secret_access_key=self.__secret_key, aws_session_token=self.__session)
        self.__savings_plan_client = boto3.client('savingsplans', aws_access_key_id=self.__access_key, aws_secret_access_key=self.__secret_key, aws_session_token=self.__session)
        self.__iam_client = boto3.client('iam', aws_access_key_id=self.__access_key, aws_secret_access_key=self.__secret_key, aws_session_token=self.__session)
        self.__assumed_role_account_name = IAMOperations(iam_client=self.__iam_client).get_account_alias_cloud_name()
        self.__cost_explorer_operations = CostExplorerOperations(ce_client=self.__ce_client)
        self.__savings_plan_operations = SavingsPlansOperations(savings_plan_client=self.__savings_plan_client)
        self.__replacement_account = literal_eval(self._environment_variables_dict.get('REPLACE_ACCOUNT_NAME'))
        self.__savings_discounts = float(self._environment_variables_dict.get('PAYER_SUPPORT_FEE_CREDIT', 0))
        self.__monthly_cost_for_spa_calc = {}
        self.__monthly_cost_for_support_fee = {}
        self.__temporary_dir = self._environment_variables_dict.get('TEMPORARY_DIR', '')

    def __get_sts_credentials(self):
        """This method returns the temporary credentials from the sts service"""
        try:
            sts_client = boto3.client("sts")
            response = sts_client.assume_role(RoleArn=self.__aws_role, RoleSessionName="PayerCostBillingReports")
            credentials = response.get('Credentials')
            return credentials.get('AccessKeyId'), credentials.get('SecretAccessKey'), credentials.get('SessionToken')
        except Exception as err:
            raise err

    def filter_data_by_tag(self, cost_data: dict, tag: str, cost_center: int = ''):
        """
        This method extract data by tag
        @param tag:
        @param cost_data: Data from the cloud explorer
        @param cost_center:
        @return: converted into dict format
        """
        data = {}
        if cost_data.get('ResultsByTime'):
            for results_by_time in cost_data.get('ResultsByTime'):
                start_time = results_by_time.get('TimePeriod').get('Start')
                for group in results_by_time.get('Groups'):
                    name = group.get('Keys')[0].split('$')[-1].strip().replace(' ', '-') if group.get('Keys') else ''
                    amount = group.get('Metrics').get('UnblendedCost').get('Amount') if group.get('Metrics') else ''
                    if name and amount:
                        if name not in data:
                            if cost_center:
                                acc_cost_center, account_budget, years, owner = self.update_to_gsheet.get_cost_center_budget_details(account_id=name, dir_path=self.__temporary_dir)
                                timestamp = datetime.datetime.strptime(start_time, '%Y-%m-%d')
                                month = datetime.datetime.strftime(timestamp, "%Y %b")
                                month_full_name = datetime.datetime.strftime(timestamp, "%B")
                                payer_monthly_savings_plan = self.update_to_gsheet.get_monthly_spa(month_name=month_full_name, dir_path=self.__temporary_dir)
                                budget = account_budget if start_time.split('-')[0] in years else 0
                                index_id = f'{start_time}-{name}'
                                upload_data = {tag: name, 'Actual': round(float(amount), self.DEFAULT_ROUND_DIGITS), 'start_date': start_time,
                                               'timestamp': timestamp, 'CloudName': 'AWS', 'Month': month,
                                               'Forecast': 0,
                                               'filter_date': f'{start_time}-{month.split()[-1]}',
                                               'Budget': round(budget / self.MONTHS, self.DEFAULT_ROUND_DIGITS), 'CostCenter': cost_center,
                                               'AllocatedBudget': budget,
                                               "Owner": owner,
                                               'SavingsPlanCost': (float(amount) / float(self.__monthly_cost_for_spa_calc.get(start_time))) * payer_monthly_savings_plan,
                                               'TotalPercentage': (float(amount) / float(self.__monthly_cost_for_spa_calc.get(start_time)))
                                               }
                                upload_data['PremiumSupportFee'] = (float(self.__monthly_cost_for_support_fee.get(start_time)) - (float(self.__monthly_cost_for_support_fee.get(start_time)) * self.__savings_discounts)) * upload_data['TotalPercentage'],
                            else:
                                index_id = f'{start_time}-{name}'
                                upload_data = {tag: name, 'Actual': round(float(amount), self.DEFAULT_ROUND_DIGITS)}
                            if index_id:
                                data[index_id] = upload_data
        if cost_data.get('DimensionValueAttributes'):
            for dimension_values in cost_data.get('DimensionValueAttributes'):
                account_id = dimension_values.get("Value")
                account = dimension_values.get('Attributes').get('description')
                if self.__replacement_account.get(account):
                    account = self.__replacement_account.get(account)
                for key_index_id in data.keys():
                    if account_id in key_index_id:
                        index_id = f'{data[key_index_id]["start_date"]}-{account}'.lower()
                        data[key_index_id].update({'Account': account, 'index_id': index_id})
        return data

    def filter_forecast_data(self, cost_forecast_data: list, cost_usage_data: dict, account_id: str, cost_center: int, account: str):
        acc_cost_center, account_budget, years, owner = self.update_to_gsheet.get_cost_center_budget_details(account_id=account_id, dir_path=self.__temporary_dir)
        for cost_forecast in cost_forecast_data:
            start_date = str((cost_forecast.get('TimePeriod').get('Start')))
            start_year = start_date.split('-')[0]
            cost = round(float(cost_forecast.get('MeanValue')), self.DEFAULT_ROUND_DIGITS)
            index = f'{start_date}-{account_id}'
            month_full_name = datetime.datetime.strftime(datetime.datetime.strptime(start_date, '%Y-%m-%d'), "%B")
            total_percentage = (cost / float(self.__monthly_cost_for_spa_calc.get(start_date)))
            payer_monthly_savings_plan = self.update_to_gsheet.get_monthly_spa(month_name=month_full_name, dir_path=self.__temporary_dir)
            if index in cost_usage_data[account]:
                cost_usage_data[account][index]['Forecast'] = cost
                cost_usage_data[account][index]['TotalPercentage'] = total_percentage
                cost_usage_data[account][index]['SavingsPlanCost'] = total_percentage * payer_monthly_savings_plan
                cost_usage_data[account][index]['PremiumSupportFee'] = total_percentage * (float(self.__monthly_cost_for_support_fee.get(start_date)) - (float(self.__monthly_cost_for_support_fee.get(start_date)) * self.__savings_discounts))
            else:
                data = {}
                data['AccountId'] = account_id
                data['Actual'] = 0
                data['Forecast'] = cost
                data['TotalPercentage'] = total_percentage
                data['SavingsPlanCost'] = total_percentage * payer_monthly_savings_plan
                data['PremiumSupportFee'] = total_percentage * (float(self.__monthly_cost_for_support_fee.get(start_date)) - (float(self.__monthly_cost_for_support_fee.get(start_date)) * self.__savings_discounts))
                data['Account'] = account
                data['start_date'] = str((cost_forecast.get('TimePeriod').get('Start')))
                data['index_id'] = f"""{data['start_date']}-{data['Account'].lower()}"""
                data['timestamp'] = datetime.datetime.strptime(data['start_date'], '%Y-%m-%d')
                data['Month'] = datetime.datetime.strftime(data['timestamp'], '%Y %b')
                data['Owner'] = owner
                if start_year in years:
                    data['Budget'] = round(account_budget / self.MONTHS, self.DEFAULT_ROUND_DIGITS)
                    data['AllocatedBudget'] = account_budget
                else:
                    data['Budget'] = 0
                    data['AllocatedBudget'] = 0
                data['CostCenter'] = cost_center
                data['CloudName'] = "AWS"
                data['filter_date'] = f'{data["start_date"]}-{data["Month"].split()[-1]}'
                cost_usage_data[account][index] = data

    def get_linked_accounts_forecast(self, linked_account_usage: dict):
        """
        This method append the forecast to the linked accounts
        """
        start_date, end_date = self.get_date_ranges(days=self.FORECAST_DAYS)
        for account, account_cost_usage in copy.deepcopy(linked_account_usage).items():
            list_usage = list(account_cost_usage.values())[0]
            cost_center, account_id = list_usage.get('CostCenter'), list_usage.get('AccountId')
            try:
                cost_forecast_data = self.__cost_explorer_operations.get_cost_forecast(start_date=start_date, end_date=end_date, granularity=self.GRANULARITY, cost_metric=self.COST_METRIC, Filter={'Dimensions': {'Key': 'LINKED_ACCOUNT', 'Values': [account_id]}})
                self.filter_forecast_data(cost_forecast_data=cost_forecast_data['ForecastResultsByTime'], cost_center=cost_center, account=account, account_id=account_id, cost_usage_data=linked_account_usage)
            except Exception as err:
                logger.info(f'No Data to get forecast: {account_id}: {account}, {err}')

    @logger_time_stamp
    def get_cost_centers(self):
        """
        This method fetch the cost centers from the account
        """
        end_date = datetime.datetime.now() - datetime.timedelta(days=5)
        start_date = end_date - datetime.timedelta(days=1)
        cost_center_data = self.__cost_explorer_operations.get_cost_and_usage_from_aws(start_date=start_date.strftime("%Y-%m-%d"), end_date=end_date.strftime("%Y-%m-%d"), granularity="MONTHLY", GroupBy=[{'Type': 'COST_CATEGORY', 'Key': 'CostCenter'}])
        return list(self.filter_data_by_tag(cost_center_data, tag='CostCenter').values())

    @logger_time_stamp
    def get_linked_accounts_usage(self):
        """This method get the linked accounts usage using cost center"""
        cost_centers = self.get_cost_centers()
        cost_usage_data = {}
        start_date, end_date = self.get_date_ranges()
        for cost_center in cost_centers:
            cost_data = self.__cost_explorer_operations.get_cost_and_usage_from_aws(start_date=start_date, end_date=end_date, granularity="MONTHLY", GroupBy=[{'Type': 'DIMENSION', 'Key': 'LINKED_ACCOUNT'}], Filter={'CostCategories': {'Key': 'CostCenter', 'Values': [cost_center.get('CostCenter')]}})
            cost_center_usage_accounts = self.filter_data_by_tag(cost_data, tag='AccountId', cost_center=int(cost_center.get('CostCenter')))
            for idx, usage in cost_center_usage_accounts.items():
                account = usage['Account']
                cost_usage_data.setdefault(account, {}).update({idx: usage})
        self.get_linked_accounts_forecast(linked_account_usage=cost_usage_data)
        self.__get_ce_cost_usage_by_filter_tag(tag_name='spot', cost_centers=cost_centers, cost_usage_data=cost_usage_data)
        handler.setLevel(logging.WARN)
        #  To prevent printing the **kwargs of the function when using the logger_time_stamp decorator.
        self.upload_data_elastic_search(linked_account_usage=cost_usage_data)
        handler.setLevel(logging.INFO)
        return cost_usage_data

    def __get_ce_cost_usage_by_filter_tag(self, cost_centers: list, tag_name: str, cost_usage_data: dict):
        """
        This method returns the cost by filter tag_name
        :param cost_centers:
        :param tag_name:
        :return:
        """
        start_date, end_date = self.get_date_ranges()
        for cost_center in cost_centers:
            cost_center_number = cost_center.get('CostCenter')
            filter_cost_center = {'CostCategories': {'Key': 'CostCenter', 'Values': [cost_center_number]}}
            values = self.__cost_explorer_operations.CE_COST_FILTERS[tag_name.upper()]['Values']
            filter_tag_value = {'Dimensions': {'Key': 'PURCHASE_TYPE', 'Values': values}}
            group_by = {'Type': 'DIMENSION', 'Key': 'LINKED_ACCOUNT'}
            cost_data = self.__cost_explorer_operations.get_cost_and_usage_from_aws(start_date=start_date, end_date=end_date, granularity="MONTHLY",
                                                                                    GroupBy=[group_by], Filter={'And': [filter_cost_center, filter_tag_value]})
            filtered_data = self.__cost_explorer_operations.get_ce_report_filter_data(ce_response=cost_data, tag_name=tag_name)
            if filtered_data:
                for index_id, row in filtered_data.items():
                    account = row.get('Account')
                    if account in self.__replacement_account:
                        account = self.__replacement_account[account]
                    if account in cost_usage_data:
                        cost_usage_data[account][index_id][f'{tag_name.title()}Usage'] = round(float(row.get(tag_name)), self.DEFAULT_ROUND_DIGITS)

    @logger_time_stamp
    def upload_data_elastic_search(self, linked_account_usage: dict):
        """This method uploads the data to elastic search"""
        for account, monthly_cost in linked_account_usage.items():
            monthly_account_cost = []
            for month, cost in monthly_cost.items():
                monthly_account_cost.append(cost)
            self.elastic_upload.es_upload_data(items=monthly_account_cost, set_index='index_id')

    def filter_cost_details_for_sp(self, total_cost: list):
        """"This method filter the account total cost"""
        results = {}
        for row in total_cost:
            start_time = row.get('TimePeriod').get('Start')
            if row.get('MeanValue'):
                cost = round(float(row.get('MeanValue')), self.DEFAULT_ROUND_DIGITS)
            else:
                cost = round(float(row.get('Total').get('UnblendedCost').get('Amount')), self.DEFAULT_ROUND_DIGITS)
            results[start_time] = cost
        return results

    def get_monthly_cost_details(self, start_date: datetime = None, end_date: datetime = None):
        """This method list the savings plan details"""
        current_date = datetime.datetime.utcnow()
        if not start_date and not end_date:
            end_date = (current_date.replace(day=1) - datetime.timedelta(days=1)).date()
            start_date = end_date.replace(day=1)
            end_date = end_date + datetime.timedelta(days=1)
        payer_cost_response = self.__cost_explorer_operations.get_cost_and_usage_from_aws(start_date=str(start_date), end_date=str(end_date), granularity='MONTHLY', cost_metric=self.COST_METRIC, Filter={'Not': {'Dimensions': {'Key': 'RECORD_TYPE', 'Values': ['Support', 'Refund', 'Credit']}}})
        payer_support_fee = self.__cost_explorer_operations.get_cost_and_usage_from_aws(start_date=str(start_date), end_date=str(end_date), granularity='MONTHLY', cost_metric=self.COST_METRIC, Filter={'Dimensions': {'Key': 'RECORD_TYPE', 'Values': ['Support']}})
        filtered_payer_cost = self.filter_cost_details_for_sp(payer_cost_response.get('ResultsByTime'))
        filtered_support_fee = self.filter_cost_details_for_sp(payer_support_fee.get('ResultsByTime'))
        self.__monthly_cost_for_spa_calc = filtered_payer_cost
        self.__monthly_cost_for_support_fee.update(filtered_support_fee)
        start_date = current_date.date()
        end_date = start_date + datetime.timedelta(days=360)
        forecast_response = self.__cost_explorer_operations.get_cost_forecast(start_date=str(start_date), end_date=str(end_date), granularity=self.GRANULARITY, cost_metric=self.COST_METRIC, Filter={'Not': {'Dimensions': {'Key': 'RECORD_TYPE', 'Values': ['Support', 'Refund', 'Credit']}}})
        payer_forecast_support_fee = self.__cost_explorer_operations.get_cost_forecast(start_date=str(start_date), end_date=str(end_date), granularity=self.GRANULARITY, cost_metric=self.COST_METRIC, Filter={'Dimensions': {'Key': 'RECORD_TYPE', 'Values': ['Support']}})
        filtered_payer_forecast = self.filter_cost_details_for_sp(forecast_response.get('ForecastResultsByTime'))
        filtered_payer_support_forecast = self.filter_cost_details_for_sp(payer_forecast_support_fee.get('ForecastResultsByTime'))
        self.__monthly_cost_for_spa_calc.update(filtered_payer_forecast)
        self.__monthly_cost_for_support_fee.update(filtered_payer_support_forecast)

    def run(self):
        """
        This method run the methods
        """
        self.get_monthly_cost_details()
        self.get_linked_accounts_usage()



import json
import logging
import os
import tempfile
from datetime import datetime

import boto3
import typeguard
from jinja2 import Environment, FileSystemLoader

from cloud_governance.common.clouds.aws.ec2.ec2_operations import EC2Operations
from cloud_governance.common.elasticsearch.elasticsearch_operations import ElasticSearchOperations
from cloud_governance.common.jira.jira_operations import JiraOperations
from cloud_governance.common.ldap.ldap_search import LdapSearch
from cloud_governance.common.logger.init_logger import handler
from cloud_governance.common.logger.logger_time_stamp import logger_time_stamp
from cloud_governance.common.mails.mail_message import MailMessage
from cloud_governance.common.mails.postfix import Postfix
from cloud_governance.main.environment_variables import environment_variables


class MonitorTickets:
    """This method monitor the Jira Tickets"""

    NEW = 'New'
    REFINEMENT = 'Refinement'
    CLOSED = 'Closed'
    IN_PROGRESS = 'In Progress'
    CLOSE_JIRA_TICKET = 0
    FIRST_CRO_ALERT: int = 5
    SECOND_CRO_ALERT: int = 3
    DEFAULT_ROUND_DIGITS: int = 3

    def __init__(self, region_name: str = ''):
        self.__environment_variables_dict = environment_variables.environment_variables_dict
        self.__cro_resource_tag_name = self.__environment_variables_dict.get('CRO_RESOURCE_TAG_NAME')
        self.__jira_operations = JiraOperations()
        self.__region_name = region_name if region_name else self.__environment_variables_dict.get('AWS_DEFAULT_REGION')
        self.es_cro_index = self.__environment_variables_dict.get('CRO_ES_INDEX', '')
        self.__default_admins = self.__environment_variables_dict.get('CRO_DEFAULT_ADMINS', [])
        self.__cloud_name = self.__environment_variables_dict.get('PUBLIC_CLOUD_NAME', '')
        self.__account = self.__environment_variables_dict.get('account', '')
        self.__es_host = self.__environment_variables_dict.get('es_host', '')
        self.__es_port = self.__environment_variables_dict.get('es_port', '')
        self.__es_operations = ElasticSearchOperations(es_host=self.__es_host, es_port=self.__es_port)
        self.__manager_escalation_days = self.__environment_variables_dict.get('MANAGER_ESCALATION_DAYS')
        self.__ldap_search = LdapSearch(self.__environment_variables_dict.get('LDAP_HOST_NAME'))
        self.__global_admin_name = self.__environment_variables_dict.get('GLOBAL_CLOUD_ADMIN')
        self.__mail_message = MailMessage()
        self.__postfix = Postfix()
        self.__ec2_operations = EC2Operations()

    @typeguard.typechecked
    @logger_time_stamp
    def get_tickets(self, ticket_status: str):
        """
        This method return the tickets based on status
        :param ticket_status:
        :return:
        """
        return self.__jira_operations.get_all_issues(ticket_status=ticket_status)

    @typeguard.typechecked
    @logger_time_stamp
    def __send_ticket_status_alerts(self, tickets: dict, ticket_status: str):
        """
        This method send alert to user Ticket status if it is on New, Refinement states
        Ticket States:
            New - Need approval
            Refinement - User didn't add the tag to the resources
        :param tickets:
        :param ticket_status:
        :return:
        """
        if ticket_status in (self.NEW, self.REFINEMENT):
            user_tickets = {}
            for ticket_id, description in tickets.items():
                ticket_id = ticket_id.split('-')[-1]
                if self.__account in description.get('AccountName'):
                    if not self.__es_operations.verify_elastic_index_doc_id(index=self.es_cro_index, doc_id=ticket_id) or ticket_id == '130':
                        if ticket_status == self.REFINEMENT:
                            ticket_status = 'manager-approved'
                        source = {'cloud_name': description.get('CloudName'), 'account_name': description.get('AccountName').replace('OPENSHIFT-', ''),
                                  'region_name': description.get('Region'), 'user': '',
                                  'user_cro': description.get('EmailAddress').split('@')[0], 'user_cost': 0, 'ticket_id': ticket_id, 'ticket_id_state': ticket_status.lower(),
                                  'estimated_cost': description.get('CostEstimation'), 'instances_count': 0, 'monitored_days': 0,
                                  'ticket_opened_date': description.get('TicketOpenedDate').date(), 'duration': description.get('Days'), 'approved_manager': '',
                                  'user_manager': '', 'project': description.get('Project'), 'owner': f'{description.get("FirstName")} {description.get("LastName")}'.upper(), 'total_spots': 0,
                                  'total_ondemand': 0, 'AllocatedBudget': [], 'instances_list': [], 'instance_types_list': []}
                        self.__es_operations.upload_to_elasticsearch(index=self.es_cro_index, data=source, id=ticket_id)
                    current_date = datetime.now().date()
                    ticket_opened_date = description.get('TicketOpenedDate').date()
                    if ticket_opened_date != current_date:
                        user = description.get('EmailAddress').split('@')[0]
                        manager = description.get('ManagerApprovalAddress').split('@')[0]
                        cc = self.__default_admins
                        subject = body = to = None
                        ticket_opened_days = (current_date - ticket_opened_date).days
                        if ticket_status == self.NEW:  # alert manager if didn't take any action
                            to = manager
                            extra_message = ''
                            if self.__manager_escalation_days <= ticket_opened_days <= self.__manager_escalation_days + 2:
                                manager_of_manager = self.__ldap_search.get_user_details(user_name=manager)
                                to = manager_of_manager.get('ManagerId', '')
                                extra_message = f"Your associate/Manager: [{manager}] doesn't approve this request.<br/>The user {user} is waiting for approval for last {ticket_opened_days} days.<br/>" \
                                                f"Please review the below details and approve/ reject"
                            elif ticket_opened_days >= self.__manager_escalation_days + 2:
                                to = self.__global_admin_name
                                extra_message = f"<b>Missing manager approval.<br />The user {user} is waiting for approval for last {ticket_opened_days} days.<br/>Please review the below details and approve/reject"
                            subject, body = self.__mail_message.cro_request_for_manager_approval(manager=to, request_user=user, cloud_name=self.__cloud_name, ticket_id=ticket_id, description=description, extra_message=extra_message)
                        else:  # alert user if doesn't add tag name
                            user_tickets.setdefault(user, []).append(f"{ticket_id} : {description.get('Project')}")
            if user_tickets:
                for user, ticket_ids in user_tickets.items():
                    active_instances = self.__ec2_operations.get_active_instances(ignore_tag='TicketId', tag_value=user, tag_name='User')
                    if active_instances:
                        for region, instances_list in active_instances.items():
                            active_instances_ids = {region: [instance.get('InstanceId') for instance in instances_list]}
                            to = user
                            cc = self.__default_admins
                            subject, body = self.__mail_message.cro_send_user_alert_to_add_tags(user=user, ticket_ids=ticket_ids)
                            with tempfile.NamedTemporaryFile(mode='w', suffix='.json') as filename:
                                filename.write(json.dumps(active_instances_ids))
                                filename.flush()
                                self.__postfix.send_email_postfix(to=to, cc=cc, subject=subject, content=body, mime_type='html', filename=filename.name)

    @typeguard.typechecked
    @logger_time_stamp
    def __track_in_progress_tickets(self, tickets: dict):
        """
        This method track the in-progress tickets
        :param tickets:
        :return:
        """
        current_date = datetime.now().date()
        for ticket_id, description in tickets.items():
            if self.__account in description.get('AccountName'):
                ticket_id = ticket_id.split('-')[-1]
                es_id_data = self.__es_operations.get_es_data_by_id(id=ticket_id, index=self.es_cro_index).get('_source')
                if es_id_data:
                    ticket_opened_date = description.get('TicketOpenedDate').date()
                    self.__region_name = description.get('Region')
                    ticket_monitoring_days = (current_date - ticket_opened_date).days
                    duration = int(es_id_data.get('duration'))
                    remaining_duration = duration - ticket_monitoring_days
                    handler.setLevel(logging.WARN)
                    es_data_change = self.verify_es_instances_state(es_data=es_id_data)
                    if es_data_change:
                        self.__es_operations.update_elasticsearch_index(index=self.es_cro_index, id=ticket_id, metadata=es_id_data)
                    self.__alert_in_progress_ticket_users(ticket_id=ticket_id, es_id_data=es_id_data, remaining_duration=remaining_duration)
                    handler.setLevel(logging.INFO)

    @typeguard.typechecked
    @logger_time_stamp
    def __tag_extend_instances(self, sub_tasks: list, ticket_id: str, duration: int, sub_task_count: int, estimated_cost: float):
        """
        This method extend the duration if the user opened expand ticket
        :param sub_tasks:
        :param ticket_id:
        :param duration:
        :param sub_task_count:
        :param estimated_cost:
        :return:
        """
        ticket_id = ticket_id.split('-')[-1]
        local_ec2_operations = EC2Operations(region=self.__region_name)
        local_ec2_client = boto3.client('ec2', region_name=self.__region_name)
        filters = {'Filters': [{'Name': f'tag:{self.__cro_resource_tag_name}', 'Values': [ticket_id]}]}
        extend_duration = 0
        for task_id in sub_tasks:
            description = self.__jira_operations.get_issue_description(ticket_id=task_id, sub_task=True)
            extend_duration += int(description.get('Days'))
        instance_ids = local_ec2_operations.get_ec2_instance_ids(**filters)
        if duration > 0:
            duration += extend_duration
            tags = [{'Key': 'Duration', 'Value': str(duration)}]
            local_ec2_operations.tag_ec2_resources(client_method=local_ec2_client.create_tags, resource_ids=instance_ids, tags=tags)
            data = {'duration': duration, 'timestamp': datetime.utcnow(), 'sub_tasks': len(sub_tasks) + sub_task_count, 'estimated_cost': round(estimated_cost, self.DEFAULT_ROUND_DIGITS)}
            if self.__es_operations:
                self.__es_operations.update_elasticsearch_index(metadata=data, id=ticket_id, index=self.es_cro_index)
            for task_id in sub_tasks:
                self.__jira_operations.move_issue_state(ticket_id=task_id, state='closed')

    @typeguard.typechecked
    @logger_time_stamp
    def __alert_in_progress_ticket_users(self, ticket_id: str, es_id_data: dict, remaining_duration: int):
        """
        This method alert the in-progress alert users
        :param ticket_id:
        :param es_id_data:
        :param remaining_duration:
        :return:
        """
        subject = body = None
        user, cc = es_id_data.get('user_cro'), self.__default_admins
        cc.append(es_id_data.get('approved_manager'))
        ticket_extended = False
        if remaining_duration <= self.FIRST_CRO_ALERT:
            sub_tasks = self.__jira_operations.get_ticket_id_sub_tasks(ticket_id=ticket_id)
            if sub_tasks:
                duration = es_id_data.get('duration')
                estimated_cost = float(es_id_data.get('estimated_cost'))
                sub_task_count = es_id_data.get('sub_tasks', 0)
                self.__tag_extend_instances(sub_tasks=sub_tasks, ticket_id=ticket_id, duration=duration, sub_task_count=sub_task_count, estimated_cost=estimated_cost)
                ticket_extended = True
        if remaining_duration == self.FIRST_CRO_ALERT:
            subject, body = self.__mail_message.cro_monitor_alert_message(user=user, days=self.FIRST_CRO_ALERT, ticket_id=ticket_id)
        elif remaining_duration == self.SECOND_CRO_ALERT:
            subject, body = self.__mail_message.cro_monitor_alert_message(user=user, days=self.SECOND_CRO_ALERT, ticket_id=ticket_id)
        else:
            if not ticket_extended:
                if remaining_duration <= self.CLOSE_JIRA_TICKET:
                    self.__update_ticket_elastic_data(ticket_id=ticket_id, es_id_data=es_id_data)
                    subject, body = self.__mail_message.cro_send_closed_alert(user, es_id_data, ticket_id)
        if subject and body:
            self.__postfix.send_email_postfix(to=user, cc=cc, subject=subject, content=body, mime_type='html')

    @typeguard.typechecked
    @logger_time_stamp
    def __update_ticket_elastic_data(self, ticket_id: str, es_id_data: dict):
        """
        This method update the ticket data in elastic_search
        :param ticket_id:
        :param es_id_data:
        :return:
        """
        self.__jira_operations.move_issue_state(ticket_id, state='CLOSED')
        data = {'timestamp': datetime.utcnow(), 'ticket_id_state': 'closed'}
        es_id_data.update(data)
        self.__es_operations.update_elasticsearch_index(index=self.es_cro_index, id=ticket_id, metadata=es_id_data)

    @typeguard.typechecked
    @logger_time_stamp
    def verify_es_instances_state(self, es_data: dict):
        """
        This method verify the state of the es_instances
        :param es_data:
        :return:
        """
        instance_ids = [resource.split(',')[1].strip() for resource in es_data.get('instances', []) if 'terminated' not in resource]
        es_data_change = False
        if instance_ids:
            local_ec2_operations = EC2Operations(region=self.__region_name)
            instances = local_ec2_operations.get_ec2_instance_ids(Filters=[{'Name': 'instance-id', 'Values': instance_ids}])
            instance_ids = list(set(instance_ids) - set(instances))
            for idx, resource in enumerate(es_data.get('instances')):
                resource_data = resource.split(',')
                instance_id = resource_data[1].strip()
                if instance_id in instance_ids:
                    es_data_change = True
                    resource_data[4] = 'terminated'
                    es_data['instances'][idx] = ', '.join(resource_data)
        return es_data_change

    @logger_time_stamp
    def __track_tickets(self):
        """
        This method trak the user tickets
        :return:
        """
        for ticket_status in [self.NEW, self.REFINEMENT]:
            self.__send_ticket_status_alerts(ticket_status=ticket_status, tickets=self.get_tickets(ticket_status=ticket_status))
        self.__track_in_progress_tickets(self.get_tickets(ticket_status=self.IN_PROGRESS))

    @logger_time_stamp
    def run(self):
        """
        This method run all methods of jira tickets monitoring
        :return:
        """
        self.__track_tickets()




from cloud_governance.policy.policy_operations.aws.zombie_non_cluster.run_zombie_non_cluster_policies import NonClusterZombiePolicy


class IpUnattached(NonClusterZombiePolicy):
    """
    Fetched the Unused elastic_ips( based on network interface Id) and delete it after 7 days of unused,
    alert user after 4 days of unused elastic_ip.
    """

    def __init__(self):
        super().__init__()

    def run(self):
        """
        This method returns zombie elastic_ip's and delete if dry_run no
        @return:
        """
        addresses = self._ec2_operations.get_elastic_ips()
        zombie_addresses = []
        for address in addresses:
            ip_no_used = False
            tags = address.get('Tags', [])
            if not self._check_cluster_tag(tags=tags) or self._get_policy_value(tags=tags) not in ('NOTDELETE', 'SKIP'):
                if not address.get('NetworkInterfaceId'):
                    ip_no_used = True
                    unused_days = self._get_resource_last_used_days(tags=tags)
                    eip_cost = self.resource_pricing.get_const_prices(resource_type='eip', hours=(self.DAILY_HOURS * unused_days))
                    delta_cost = 0
                    if unused_days == self.DAYS_TO_NOTIFY_ADMINS:
                        delta_cost = self.resource_pricing.get_const_prices(resource_type='eip', hours=(self.DAILY_HOURS * (unused_days - self.DAYS_TO_TRIGGER_RESOURCE_MAIL)))
                    else:
                        if unused_days >= self.DAYS_TO_DELETE_RESOURCE:
                            delta_cost = self.resource_pricing.get_const_prices(resource_type='eip', hours=(self.DAILY_HOURS * (unused_days - self.DAYS_TO_NOTIFY_ADMINS)))
                    zombie_eip = self._check_resource_and_delete(resource_name='ElasticIp',
                                                                 resource_id='AllocationId',
                                                                 resource_type='AllocateAddress',
                                                                 resource=address,
                                                                 empty_days=unused_days,
                                                                 days_to_delete_resource=self.DAYS_TO_DELETE_RESOURCE, tags=tags,
                                                                 extra_purse=eip_cost, delta_cost=delta_cost)
                    if zombie_eip:
                        zombie_addresses.append({'ResourceId': address.get('AllocationId'),
                                                 'Name': self._get_tag_name_from_tags(tags=tags),
                                                 'User': self._get_tag_name_from_tags(tags=tags, tag_name='User'),
                                                 'PublicIp': address.get('PublicIp'),
                                                 'Skip': self._get_policy_value(tags=tags),
                                                 'Days': unused_days})
                else:
                    unused_days = 0
                self._update_resource_tags(resource_id=address.get('AllocationId'), tags=tags, left_out_days=unused_days, resource_left_out=ip_no_used)
        return zombie_addresses



import datetime
from ast import literal_eval

import boto3

from cloud_governance.common.clouds.aws.cloudtrail.cloudtrail_operations import CloudTrailOperations
from cloud_governance.common.clouds.aws.iam.iam_operations import IAMOperations
from cloud_governance.common.clouds.aws.ec2.ec2_operations import EC2Operations
from cloud_governance.common.clouds.aws.price.resources_pricing import ResourcesPricing
from cloud_governance.common.clouds.aws.s3.s3_operations import S3Operations
from cloud_governance.common.elasticsearch.elastic_upload import ElasticUpload
from cloud_governance.common.ldap.ldap_search import LdapSearch
from cloud_governance.common.logger.init_logger import logger
from cloud_governance.common.mails.mail_message import MailMessage
from cloud_governance.common.mails.postfix import Postfix
from cloud_governance.main.environment_variables import environment_variables
from cloud_governance.policy.aws.zombie_cluster_resource import ZombieClusterResources


class NonClusterZombiePolicy:

    DAYS_TO_DELETE_RESOURCE = 7
    DAYS_TO_NOTIFY_ADMINS = 6
    DAYS_TO_TRIGGER_RESOURCE_MAIL = 4
    DAILY_HOURS = 24

    def __init__(self):
        self.__environment_variables_dict = environment_variables.environment_variables_dict
        self._end_date = datetime.datetime.now()
        self._start_date = self._end_date - datetime.timedelta(days=self.DAYS_TO_DELETE_RESOURCE)
        self._account = self.__environment_variables_dict.get('account', '')
        self._dry_run = self.__environment_variables_dict.get('dry_run', 'yes')
        self._region = self.__environment_variables_dict.get('AWS_DEFAULT_REGION', 'us-east-2')
        self._policy = self.__environment_variables_dict.get('policy', '')
        self._policy_output = self.__environment_variables_dict.get('policy_output', '')
        self._ec2_client = boto3.client('ec2', region_name=self._region)
        self._ec2_operations = EC2Operations(region=self._region)
        self._s3_client = boto3.client('s3')
        self._iam_operations = IAMOperations()
        self._iam_client = boto3.client('iam')
        self._cluster_prefix = 'kubernetes.io/cluster'
        self._zombie_cluster = ZombieClusterResources(cluster_prefix=self._cluster_prefix)
        self._s3operations = S3Operations(region_name=self._region)
        self._cloudtrail = CloudTrailOperations(region_name=self._region)
        self._special_user_mails = self.__environment_variables_dict.get('special_user_mails', '{}')
        self._account_admin = self.__environment_variables_dict.get('account_admin', '')
        self._mail = Postfix()
        self._mail_description = MailMessage()
        self.__ldap_host_name = self.__environment_variables_dict.get('LDAP_HOST_NAME', '')
        self._ldap = LdapSearch(ldap_host_name=self.__ldap_host_name)
        self.__email_alert = self.__environment_variables_dict.get('EMAIL_ALERT') if self.__environment_variables_dict.get('EMAIL_ALERT') else False
        self.__manager_email_alert = self.__environment_variables_dict.get('MANAGER_EMAIL_ALERT')
        self._admins = ['athiruma@redhat.com', 'ebattat@redhat.com']
        self._es_upload = ElasticUpload()
        self.resource_pricing = ResourcesPricing()

    def set_dryrun(self, value: str):
        self._dry_run = value

    def set_policy(self, value: str):
        self._policy = value

    def set_region(self, value: str):
        self._region = value

    def _literal_eval(self, data: any):
        tags = {}
        if data:
            tags = literal_eval(data)
        return tags

    def _check_cluster_tag(self, tags: list):
        """
        This method returns True if it is live cluster tag is active False not
        @param tags:
        @return:
        """
        if tags:
            for tag in tags:
                if tag.get('Key').startswith(self._cluster_prefix):
                    return True
        return False

    def _get_tag_name_from_tags(self, tags: list, tag_name: str = 'Name'):
        """
        This method returns the tag_name from resource tags
        @param tags:
        @param tag_name:
        @return:
        """
        if tags:
            for tag in tags:
                if tag.get('Key').strip().lower() == tag_name.lower():
                    return tag.get('Value').strip()
            return ''
        return ''

    def _calculate_days(self, create_date: datetime):
        """
        This method returns the days
        @return:
        """
        today = datetime.date.today()
        days = today - create_date.date()
        return days.days

    def _get_ami_ids(self):
        """
        This method returns all image ids
        @return:
        """
        images = self._ec2_operations.get_images()
        image_ids = []
        for image in images:
            image_ids.append(image.get('ImageId'))
        return image_ids

    def _beautify_upload_data(self, upload_resource_data: list):
        """
        This method beautify the data to upload to elasticsearch
        @param upload_resource_data:
        @return:
        """
        upload_data = []
        for resource_data in upload_resource_data:
            if isinstance(resource_data, list):
                resource_str = ''
                for resource in resource_data[:-1]:
                    resource_str += f'{resource} | '
                resource_str += str(resource_data[-1])
                upload_data.append(resource_str)
            else:
                upload_data.append(resource_data)
        return upload_data

    def _get_resource_username(self, resource_id: str, resource_type: str, create_date: datetime = '', event_type: str = ''):
        """
        Get Username from the cloudtrail
        @param create_date:
        @param resource_id:
        @param resource_type:
        @return:
        """
        if event_type:
            return self._cloudtrail.get_username_by_instance_id_and_time(start_time=create_date, resource_id=resource_id,
                                                                         resource_type=resource_type, event_type=event_type)
        else:
            return self._cloudtrail.get_username_by_instance_id_and_time(start_time=create_date, resource_id=resource_id,
                                                                         resource_type=resource_type)

    def _get_policy_value(self, tags: list):
        """
        This method beautify the value
        @param tags:
        @return:
        """
        policy_value = self._get_tag_name_from_tags(tags=tags, tag_name='Policy').strip()
        if policy_value:
            return policy_value.replace('_', '').replace('-', '').upper()
        return 'NA'

    def _trigger_mail(self, tags: list, resource_id: str, days: int, resource_type: str, **kwargs):
        """
        This method send triggering mail
        @param tags:
        @param resource_id:
        @return:
        """
        if self.__email_alert:
            try:
                special_user_mails = self._literal_eval(self._special_user_mails)
                user, resource_name = self._get_tag_name_from_tags(tags=tags, tag_name='User'), self._get_tag_name_from_tags(
                    tags=tags, tag_name='Name')
                if not resource_name:
                    resource_name = self._get_tag_name_from_tags(tags=tags, tag_name='cg-Name')
                to = user if user not in special_user_mails else special_user_mails[user]
                ldap_data = self._ldap.get_user_details(user_name=to)
                cc = [self._account_admin, f'{ldap_data.get("managerId")}@redhat.com'] if self.__manager_email_alert else []
                name = to
                if ldap_data:
                    name = ldap_data.get('displayName')
                subject, body = self._mail_description.resource_message(name=name, days=days,
                                                                        notification_days=self.DAYS_TO_TRIGGER_RESOURCE_MAIL,
                                                                        delete_days=self.DAYS_TO_DELETE_RESOURCE,
                                                                        resource_name=resource_name, resource_id=resource_id,
                                                                        resource_type=resource_type, msgadmins=self.DAYS_TO_NOTIFY_ADMINS, extra_purse=kwargs.get('extra_purse'))
                if not kwargs.get('admins'):
                    self._mail.send_email_postfix(to=to, content=body, subject=subject, cc=cc, resource_id=resource_id, message_type=kwargs.get('message_type'), extra_purse=kwargs.get('delta_cost', 0))
                else:
                    if self.__manager_email_alert:
                        kwargs['admins'].append(f'{ldap_data.get("managerId")}@redhat.com')
                    self._mail.send_email_postfix(to=kwargs.get('admins'), content=body, subject=subject, cc=[], resource_id=resource_id, message_type=kwargs.get('message_type'), extra_purse=kwargs.get('delta_cost', 0))
            except Exception as err:
                logger.info(err)

    def _update_tag_value(self, tags: list, tag_name: str, tag_value: str):
        """
        This method updates the tag value
        @param tags:
        @param tag_name:
        @param tag_value:
        @return:
        """
        found = False
        if tags:
            for tag in tags:
                if tag.get('Key') == tag_name:
                    tag['Value'] = str(tag_value)
                    found = True
            if not found:
                tags.append({'Key': tag_name, 'Value': str(tag_value)})
            return tags
        return [{'Key': tag_name, 'Value': str(tag_value)}]

    def _get_resource_last_used_days(self, tags: list):
        """
        This method get last used day from the tags
        @param tags:
        @return:
        """
        last_used_day = self._get_tag_name_from_tags(tags=tags, tag_name='LastUsedDay')
        if not last_used_day:
            last_used_day = 1
        else:
            last_used_day = int(last_used_day) + 1
        return last_used_day

    def __delete_resource_on_name(self, resource_id: str):
        try:
            if self._policy == 's3_inactive':
                self._s3_client.delete_bucket(Bucket=resource_id)
            elif self._policy == 'empty_roles':
                self._iam_client.delete_role(RoleName=resource_id)
            elif self._policy == 'ebs_unattached':
                self._ec2_client.delete_volume(VolumeId=resource_id)
            elif self._policy == 'ip_unattached':
                self._ec2_client.release_address(AllocationId=resource_id)
            elif self._policy == 'unused_nat_gateway':
                self._ec2_client.delete_nat_gateway(NatGatewayId=resource_id)
            elif self._policy == 'zombie_snapshots':
                self._ec2_client.delete_snapshot(SnapshotId=resource_id)
            logger.info(f'{self._policy} deleted: {resource_id}')
        except Exception as err:
            logger.info(f'Exception raised: {err}: {resource_id}')

    def _check_resource_and_delete(self, resource_name: str, resource_id: str, resource_type: str, resource: dict, empty_days: int, days_to_delete_resource: int, tags: list = [], **kwargs):
        """
        This method check and delete resources
        @param resource_name:
        @param resource_id:
        @param resource_type:
        @param resource:
        @param empty_days:
        @param days_to_delete_resource:
        @param tags:
        @return:
        """
        resource_id = resource.get(resource_id)
        if not tags:
            tags = resource.get('Tags') if resource.get('Tags') else []
        user = self._get_tag_name_from_tags(tag_name='User', tags=tags)
        if not user:
            user = self._get_resource_username(resource_id=resource_id, resource_type=resource_type, event_type='EventName')
            if user:
                tags.append({'Key': 'User', 'Value': user})
        zombie_resource = {}
        if empty_days >= self.DAYS_TO_TRIGGER_RESOURCE_MAIL:
            if empty_days == self.DAYS_TO_TRIGGER_RESOURCE_MAIL:
                kwargs['delta_cost'] = kwargs.get('extra_purse')
                self._trigger_mail(resource_type=resource_name, resource_id=resource_id, tags=tags, days=self.DAYS_TO_TRIGGER_RESOURCE_MAIL, message_type='notification', extra_purse=kwargs.get('extra_purse'), delta_cost=kwargs.get('delta_cost', 0))
            elif empty_days == self.DAYS_TO_NOTIFY_ADMINS:
                self._trigger_mail(resource_type=resource_name, resource_id=resource_id, tags=tags, days=empty_days, admins=self._admins, message_type='notify_admin', extra_purse=kwargs.get('extra_purse'), delta_cost=kwargs.get('delta_cost', 0))
            elif empty_days >= days_to_delete_resource:
                if self._dry_run == 'no':
                    if self._get_policy_value(tags=tags) not in ('NOTDELETE', 'SKIP'):
                        self._trigger_mail(resource_type=resource_name, resource_id=resource_id, tags=tags, days=empty_days, message_type='delete', extra_purse=kwargs.get('extra_purse'), delta_cost=kwargs.get('delta_cost', 0))
                        self.__delete_resource_on_name(resource_id=resource_id)
            zombie_resource = resource
        return zombie_resource

    def _update_resource_tags(self, resource_id: str, left_out_days: int, tags: list, resource_left_out: bool):
        """
        This method update the tags in aws
        @return:
        """
        if left_out_days < 7 or self._dry_run == 'yes' or self._get_policy_value(tags=tags) in ('NOTDELETE', 'SKIP'):
            if self._get_tag_name_from_tags(tags=tags, tag_name='LastUsedDay') or resource_left_out:
                tags = self._update_tag_value(tags=tags, tag_name='LastUsedDay', tag_value=str(left_out_days))
                try:
                    if self._policy == 's3_inactive':
                        self._s3_client.put_bucket_tagging(Bucket=resource_id, Tagging={'TagSet': tags})
                    elif self._policy == 'empty_roles':
                        self._iam_client.tag_role(RoleName=resource_id, Tags=tags)
                    elif self._policy in ('ip_unattached', 'unused_nat_gateway', 'zombie_snapshots'):
                        self._ec2_client.create_tags(Resources=[resource_id], Tags=tags)
                except Exception as err:
                    logger.info(f'Exception raised: {err}: {resource_id}')

    def _organise_instance_data(self, resources: list):
        """
        This method convert all datetime into string
        @param resources:
        @return:
        """
        organize_data = []
        if 'ec2' in self._policy:
            for instance in resources:
                instance['LaunchTime'] = instance['LaunchTime'].strftime("%Y-%m-%dT%H:%M:%S+00:00")
                for index, device_mappings in enumerate(instance['BlockDeviceMappings']):
                    instance['BlockDeviceMappings'][index]['Ebs']['AttachTime'] = device_mappings['Ebs']['AttachTime'].strftime("%Y-%m-%dT%H:%M:%S+00:00")
                for index, network_interface in enumerate(instance['NetworkInterfaces']):
                    instance['NetworkInterfaces'][index]['Attachment']['AttachTime'] = network_interface['Attachment']['AttachTime'].strftime("%Y-%m-%dT%H:%M:%S+00:00")
                if instance.get('UsageOperationUpdateTime'):
                    instance['UsageOperationUpdateTime'] = instance['UsageOperationUpdateTime'].strftime("%Y-%m-%dT%H:%M:%S+00:00")
                if instance.get('metrics'):
                    for index, metric in enumerate(instance['metrics']):
                        instance['metrics'][index]['Timestamps'] = [date.strftime("%Y-%m-%dT%H:%M:%S+00:00") for date in metric['Timestamps']]
                organize_data.append(instance)
        else:
            for volume in resources:
                if volume.get('Attachments'):
                    for attachment in volume.get('Attachments'):
                        attachment['AttachTime'] = attachment['AttachTime'].strftime("%Y-%m-%dT%H:%M:%S+00:00")
                volume['CreateTime'] = volume['CreateTime'].strftime("%Y-%m-%dT%H:%M:%S+00:00")
                organize_data.append(volume)
        return organize_data

    def get_ebs_cost(self, resource: any, resource_type: str, resource_hours: int):
        ebs_cost = 0
        if resource_type == 'ec2':
            volume_ids = []
            for block_device in resource:
                volume_ids.append(block_device.get('Ebs').get('VolumeId'))
            volumes = self._ec2_client.describe_volumes(VolumeIds=volume_ids)['Volumes']
            for volume in volumes:
                ebs_cost += self.resource_pricing.get_ebs_cost(volume_size=volume.get('Size'), volume_type=volume.get('VolumeType'), hours=resource_hours)
        else:
            if resource_type == 'ebs':
                ebs_cost += self.resource_pricing.get_ebs_cost(volume_size=resource.get('Size'), volume_type=resource.get('VolumeType'), hours=resource_hours)
        return round(ebs_cost, 3)



import os.path

from jinja2 import Environment, FileSystemLoader

from cloud_governance.common.ldap.ldap_search import LdapSearch
from cloud_governance.main.environment_variables import environment_variables


class MailMessage:

    RESTRICTION = 'Do not reply this email. If you need more information, please reach out to us in the slack channel - #perf-dept-public-clouds.'
    FOOTER = '<div style="color:gray">---<br />Thanks, <br />Cloud GovernanceTeam</div>'

    def __init__(self):
        self.__environment_variables_dict = environment_variables.environment_variables_dict
        self.account = self.__environment_variables_dict.get('account', '').upper()
        self.policy = self.__environment_variables_dict.get('policy', '')
        self.region = self.__environment_variables_dict.get('AWS_DEFAULT_REGION', '')
        self.__LDAP_HOST_NAME = self.__environment_variables_dict.get('LDAP_HOST_NAME')
        self.__ldap_search = LdapSearch(ldap_host_name=self.__LDAP_HOST_NAME)
        self.__public_cloud_name = self.__environment_variables_dict.get('PUBLIC_CLOUD_NAME')
        self.__portal = self.__environment_variables_dict.get('CRO_PORTAL', '')
        self.__cro_duration_days = self.__environment_variables_dict.get('CRO_DURATION_DAYS')
        self.__LDAP_HOST_NAME = self.__environment_variables_dict.get('LDAP_HOST_NAME')
        self.__ldap_search = LdapSearch(ldap_host_name=self.__LDAP_HOST_NAME)
        self.__templates_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'templates')
        self.env_loader = Environment(loader=FileSystemLoader(self.__templates_path))

    def get_user_ldap_details(self, user_name: str):
        """
        This method return user details from ldap
        :param user_name:
        :return:
        """
        user_details = self.__ldap_search.get_user_details(user_name=user_name)
        if user_details:
            return user_details.get('displayName')
        else:
            return None

    def ec2_stop(self, name: str, days: int, image_id: str, delete_instance_days: int, instance_name: str,
                 resource_id: str, stopped_time: str, ec2_type: str, **kwargs):
        extra_purse = ''
        if kwargs.get("extra_purse"):
            extra_purse = f', Cost {round(kwargs.get("extra_purse"), 3)} $ '
        subject = f'cloud-governance alert: ec2-stop'
        content = 'If you do not want it to be deleted, please add "Policy=Not_Delete" or "Policy=skip" tag to this instance.'
        message = f'This instance will be deleted in the {delete_instance_days-days} days if no further action is taken.'
        if kwargs.get('msgadmins'):
            message = f'This instance will be deleted in the {delete_instance_days - days} days if no further action is taken.'
        elif image_id != '':
            content = f'You can find a image of the deleted instance under AMI: {image_id}'
            message = f'This instance will be deleted due to it was in stopped state more than {delete_instance_days} days.'
        body = f"""
Hi {name},

Instance: {instance_name}: {resource_id}( InstanceType:{ec2_type}{extra_purse}) in {self.region} region in account: {self.account} was stopped on {stopped_time}, it stopped state more than {days} days.
{message}
{content}

{self.RESTRICTION}

Best regards,
Cloud-governance Team""".strip()
        return subject, body

    def ec2_idle(self, name: str, days: int, notification_days: int, stop_days: int, instance_name: str,
                 resource_id: str, ec2_type: str, **kwargs):
        extra_purse = ''
        if kwargs.get('extra_purse'):
            extra_purse = f', Cost {round(kwargs.get("extra_purse"), 3)} $ '
        subject = f'cloud-governance alert: ec2-idle'
        if days == notification_days:
            cause = f'This instance will be stopped in {stop_days-days} days if no further action is taken'
            content = 'If you do not want it to be deleted, please "Policy=Not_Delete" or "Policy=ski"p tag to this instance.'
        else:
            cause = f'This instance will be stopped.'
            content = 'In future cloud-governance will not stop it add "Policy=Not_Delete" or "Policy=skip" tag to your instance'
        body = f"""
Hi {name},

Instance: {instance_name}: {resource_id} ( InstanceType:{ec2_type}{extra_purse}) in {self.region} on account: {self.account} is idle more than {days} days.
{cause}
{content}
If you already added the tag, please ignore this mail.

{self.RESTRICTION}

Best Regards,
Cloud-governance Team""".strip()
        return subject, body

    def iam_user_add_tags(self, name: str, user: str, spreadsheet_id: str):
        subject = f'cloud-governance alert: Missing tags in AWS IAM User'
        body = f"""
Hi {name},

{self.__environment_variables_dict.get('account')} IAM User: {user} has missing tags 
Please add the tags in the spreadsheet: https://docs.google.com/spreadsheets/d/{spreadsheet_id}.
If you already filled the tags, please ignore the mail.

{self.RESTRICTION}

Best Regards
Cloud-governance Team""".strip()
        return subject, body

    def aws_user_over_usage_cost(self, user: str, usage_cost: int, name: str, user_usage: int):
        """
        This method send subject, body to over usage cost
        @return:
        """
        subject = f'cloud-governance alert: last week cost over usage {usage_cost}$ per user'
        body = f"""
Hi {name},
Your AWS user:{user} in account:{self.__environment_variables_dict.get('account')} cost was {user_usage}$ in the last week. 
Please verify that you are using all the resources.
Please verify that you are using all the resources in attached {user}_resource.json file.

{self.RESTRICTION}

Best Regards,
Cloud-governance Team""".strip()
        return subject, body

    def resource_message(self, name: str, days: int, notification_days: int, delete_days: int, resource_name: str, resource_id: str, resource_type: str, resources: list = [], **kwargs):
        """
        This method prepare mail message based on resource_type and return subject, body
        @param resources:
        @param name:
        @param days:
        @param notification_days:
        @param delete_days:
        @param resource_name:
        @param resource_id:
        @param resource_type:
        @return:
        """
        resource_type = resource_type.capitalize()
        reason = self.policy.split('_')[-1]
        extra_purse = ''
        if kwargs.get('extra_purse'):
            extra_purse = f'(Cost {round(kwargs.get("extra_purse"), 3)} $)'
        if 'empty' in self.policy:
            reason = 'empty'
        if 'zombie' in self.policy:
            reason = 'unused'
        subject = f'cloud-governance alert: {self.policy}'
        if days == notification_days:
            cause = f'This {resource_type} will be deleted in {delete_days-days} days if no further action is taken'
            content = f'If you do not want to be deleted, please add "Policy=Not_Delete" or "Policy=skip" tag to this {resource_type}.'
        elif kwargs.get('msgadmins'):
            cause = f'This {resource_type} will be deleted in {delete_days - kwargs.get("msgadmins")} days if no further action is taken'
            content = f'If you do not want to be deleted, please add "Policy=Not_Delete" or "Policy=skip" tag to this {resource_type}.'
        else:
            cause = f'This {resource_type} will be deleted due to it was {reason} more than {delete_days} days.'
            content = f'In future cloud-governance will not delete your {resource_type} add "Policy=Not_Delete" or "Policy=skip" tag to your {resource_type}s'
        body = f"""
Hi {name},

{resource_type.upper()}: {resource_name}: {resource_id} {extra_purse} in {self.region} region in account: {self.account} has been in {reason} for more than {days} days.
{cause}
{content}
If you already added the tag, please ignore this mail.

{self.RESTRICTION}

Best Regards,
Cloud-governance Team""".strip()
        return subject, body

    def zombie_cluster_mail_message(self, name: str, days: int, notification_days: int, resource_name: str, delete_days: int):
        """
        This method prepare mail message based on resource_type and return subject, body
        @param delete_days:
        @param name:
        @param days:
        @param notification_days:
        @param resource_name:
        @return:
        """
        subject = f'Cloud-governance policy alert: {self.policy}'
        if days == notification_days:
            cause = f'Cloud-governance will delete those {self.policy}s in the {delete_days-days} days.'
        else:
            cause = f'Those {self.policy}s will be deleted.'
        body = f"""
Hi {name},

There are zombie resources that related to your terminated cluster: {resource_name}, region: {self.region}, AWS account: {self.account}
You can find your {self.policy}s in the attached file: {resource_name.replace('/', '-')}.json
{cause}

{self.RESTRICTION}

Best Regards,
Cloud-governance Team""".strip()
        return subject, body

    def monthly_html_mail_message(self, data: str):
        """
        This method returns the description of the monthly resources
        @return:
        """
        subject = 'Cloud-Governance Monthly Report'
        body = f"""
<div class="greeting">
    <p>Hi All,</p>
</div>

<div class="body">
    <p>You can find below <a href="https://github.com/redhat-performance/cloud-governance" style="text-decoration:none;">cloud-governance</a> summary monthly report:</p>
    {data}
    <p>For more details open Cloud-Governance <a href="http://grafana.intlab.perf-infra.lab.eng.rdu2.redhat.com/" style="text-decoration:none;" target="_blank">Grafana</a>. [ user/password: cloud_governance ]</p>
    <p>Do not reply to this email, in case of any further questions. Please reach out to us in the slack channel - <span style="color:red">#perf-dept-public-clouds.<span>'</p>
</div>
<div style="color:gray" class="footer">
    <address>
        --<br/>
        Best Regards,<br/>
        Cloud-governance Team<br/>
    </address>
</div>
""".strip()
        return subject, body

    def ibm_cost_over_usage(self, data: str, month: int, year: int):
        """
        This method returns the mail message of cost over usage
        @return:
        """
        subject = 'Cloud-Governance IBM Cost Over Usage'
        body = f"""
        <div class="greeting">
            <p>Hi,</p>
        </div>
        
        <div class="body">
            <p>This following services were over limit</p>
            {data}
            <p>For more details open IBM Billing <a href="https://cloud.ibm.com/billing/usage?month={year}-{month}/" style="text-decoration:none;" target="_blank">Console</a>.</p>
            <p>Do not reply to this email, in case of any further questions. Please reach out to us in the slack channel -  <span style="color:red">#perf-dept-public-clouds.<span>'</p>
        </div>
        <div>
            <img src="https://upload.wikimedia.org/wikipedia/commons/2/24/IBM_Cloud_logo.png"/>
        </div>
        <div style="color:gray" class="footer">
            <address>
                --<br/>
                Best Regards,<br/>
                Cloud-governance Team<br/>
            </address>
        </div>
        """.strip()
        return subject, body

    def cro_monitor_alert_message(self, days: int, user: str, ticket_id: str):
        """
        This method return the CRO Alert Message
        :param days:
        :param user:
        :param ticket_id:
        :return:
        """
        ticket_id = ticket_id.split('-')[-1]
        subject = f'[Action required] Cloud Resources Budget request Ticket Expiring in {days} days'
        user_display_name = self.get_user_ldap_details(user_name=user)
        body = f"""
                <div>
                    <p>Hi {user_display_name},</p>
                </div>
                <div>
                    <p>You project budget request ( TicketId: {ticket_id} ) will be expired in {days} days.</p>
                    <p>You can extend the project duration in the following url {self.__portal} or terminate the instances</p>
                    <p>Visit the <a href="{self.__portal}">wiki page</a> to get more information</p>
                </div>
                {self.FOOTER}
""".strip()
        return subject, body

    def cro_cost_over_usage(self, **kwargs):
        """
        This method returns the subject, body for cost over usage
        :param kwargs:
        :return:
        """
        cloud_name = kwargs.get('CloudName', 'NA').upper()
        over_usage_cost = kwargs.get('OverUsageCost', 'NA')
        full_name = kwargs.get('FullName', '')
        if not full_name:
            full_name = kwargs.get('to')
        user_cost = round(kwargs.get('Cost', 0), 3)
        subject = f' [Action required]: Cloud Resources Open Budget Request'
        if user_cost > over_usage_cost:
            message = f"it's over $ {over_usage_cost}."
        else:
            message = f"it may over $ {over_usage_cost} in next few days."
        body = f"""
        <div>
        Hi {full_name},
        </div><br/>
        <div>
            Your {cloud_name} cost usage in the last {self.__cro_duration_days} days is $ {user_cost} and {message}<br/>
            You must open the project ticket in the following <a href="{self.__portal}">Link</a>.<br />
            After submitting a ticket, you must add Tag (TicketId:#) to every active resource that is related to the project ticket.<br/>
            
            If you have any questions, please let us know in slack channel #perf-dept-public-clouds
        <div><br/><br/>
        {self.FOOTER}
"""
        return subject, body

    def cro_request_for_manager_approval(self, manager: str, request_user: str, cloud_name: str, ticket_id: str, description: dict, **kwargs):
        """
        This method returns the message for manager, regarding user approval
        :param description:
        :param ticket_id:
        :param manager:
        :param request_user:
        :param cloud_name:
        :return:
        """
        subject = '[Action required]: Cloud Resources Budget Request Approval'
        manager_full_name = self.get_user_ldap_details(user_name=manager)
        user_full_name = self.get_user_ldap_details(user_name=request_user)
        ticket_id = ticket_id.split('-')[-1]
        context = {'manager': manager, 'manager_full_name': manager_full_name, 'user_full_name': user_full_name,
                   'ticket_id': ticket_id, 'portal': self.__portal, 'request_user': request_user, 'description': description,
                   'footer': self.FOOTER}
        template_loader = self.env_loader.get_template('cro_request_for_manager_approval.j2')
        context['extra_message'] = kwargs.get('extra_message', '')
        body = template_loader.render(context)
        return subject, body

    def cro_send_user_alert_to_add_tags(self, user: str, ticket_ids: list):
        """
        This method return the subject, body for adding tags
        :param user:
        :param ticket_ids:
        :return:
        """
        subject = '[Action required]: Add TicketId tag'
        ticket_ids = "\n".join([f"<li>{val}</li>" for idx, val in enumerate(ticket_ids)])
        user_display_name = self.get_user_ldap_details(user_name=user)
        body = f"""
        <div>Hi {user_display_name},</div>
        <p>You have the following <b>Approved</b> JIRA Ticket-Ids</p>
        <ul>{ticket_ids}</ul><br />
        Currently, there are several instances running over budget, kindly review and tag instances with TicketId: #</p>
        <br />Please find the below attached document.<br />
        </div><br />
        {self.FOOTER}
        """
        return subject, body

    def cro_send_closed_alert(self, user: str, es_data: dict, ticket_id: str):
        """
        This method send cro ticket close alert
        :param user:
        :param es_data:
        :param ticket_id:
        :return:
        """
        subject = 'Closing Cloud Budget Request ticket'
        ticket_id = ticket_id.split('-')[-1]
        user_full_name = self.get_user_ldap_details(user_name=user)
        body = f"""
        <div>Hi {user_full_name},</div><br />
            <div>
            Your cloud budget request ( TicketId: {ticket_id} ) duration expired and the ticket auto closed.<br />
            You can find the summary in <a href="{self.__portal}/wiki/clouds">Portal</a>.<br />
            </div><br /><br/>
        {self.FOOTER}
        """
        return subject, body

    def filter_resources_on_days(self, resources: dict):
        """
        This method return the resources based on the days
        :param resources:
        :param days:
        :return:
        """
        resources_by_days = {}
        for policy_name, resource_data in resources.items():
            for region_name, policy_region_data in resource_data.items():
                for data_item in policy_region_data:
                    resources_by_days.setdefault(data_item.get('Days'), []).append(data_item)
        return resources_by_days

    def get_data_in_html_table_format(self, resources: dict):
        """
        This method return user policy alerts in HTML table format
        :param resources:
        :return:
        """
        style = """
                    <style>
                    #customers {
                    font-family: Verdana, Helvetica, sans-serif;
                    border-collapse: collapse;
                    width: 100%;
                    }

                    #customers td, #customers th {
                    border: 2px solid #000;
                    padding: 8px;
                    align: left;
                    }

                    #customers tr:nth-child(even){background-color: #dddddd;}

                    #customers tr:hover {background-color: #B9D9B7;}

                    #customers th {
                    padding-top: 12px;
                    padding-bottom: 12px;
                    text-align: left;
                    background-color: #04AA6D;
                    color: white;
                    }
                    </style>
                """
        html_table_format = f"""{style}<table id="customers">"""
        thead_values = ['Policy', 'Region', 'ResourceId', 'Name', 'Action', 'DeletedDay']
        th_elements = ''.join([f'<th>{value}</th>' for value in thead_values])
        html_table_format += f'<thead><tr>{th_elements}</tr></thead><tbody>'
        for days, resource_data in resources.items():
            resource_data = sorted(resource_data, key=lambda item: (item.get('Policy'), item.get('Region')))
            for resource in resource_data:
                html_table_format += '<tr>'
                for th_value in thead_values:
                    if 'Deleted' == resource.get(th_value):
                        html_table_format += f"<td>{resource.get(th_value)} &#128465;</td>"
                    else:
                        html_table_format += f"""<td>{resource.get(th_value)}</td>"""
                html_table_format += '</tr>'
        html_table_format += '</tbody></table>'
        return html_table_format

    def get_agg_policies_mail_message(self, user: str, user_resources: dict):
        """
        This method return the message for the aggregated alert of all policies
        :param user:
        :param user_resources:
        :return:
        """
        display_name = self.get_user_ldap_details(user_name=user)
        resources_by_days = self.filter_resources_on_days(resources=user_resources)
        table_data = self.get_data_in_html_table_format(resources=resources_by_days)
        display_name = display_name if display_name else user
        subject = f'Cloud Governance: Policy Alerts'
        body = f"""
        <div>
            <p>Hi {display_name},</p>
        </div>
        <div>
            <p>You can find below your unused resources in the {self.__public_cloud_name} account ({self.account}).</p>
            <p>If you want to keep them, please add "Policy=Not_Delete" or "Policy=skip" tag for each resource</p>
            {table_data}
        </div>
        <p>{self.RESTRICTION}</p>
        {self.FOOTER}
"""
        return subject, body



import logging
from datetime import datetime, timedelta

import typeguard

from cloud_governance.cloud_resource_orchestration.aws.ec2.cost_over_usage import CostOverUsage
from cloud_governance.cloud_resource_orchestration.aws.ec2.monitor_tickets import MonitorTickets
from cloud_governance.common.clouds.aws.iam.iam_operations import IAMOperations
from cloud_governance.common.jira.jira_operations import JiraOperations
from cloud_governance.common.logger.init_logger import handler
from cloud_governance.common.logger.logger_time_stamp import logger_time_stamp
from cloud_governance.main.environment_variables import environment_variables


class CollectCROReports:
    """
    This method collects the user/instance-id data from the cost-explorer
    """

    DEFAULT_ROUND_DIGITS = 3
    ZERO = 0
    TICKET_ID_KEY = 'ticket_id'
    COST_EXPLORER_TAGS = {TICKET_ID_KEY: 'TicketId'}
    AND = 'And'
    ALLOCATED_BUDGET = 'AllocatedBudget'

    def __init__(self):
        self.__environment_variables_dict = environment_variables.environment_variables_dict
        self.__account_name = self.__environment_variables_dict.get('account', '').replace('OPENSHIFT-', '').strip()
        self.__cost_over_usage = CostOverUsage()
        self.jira_operations = JiraOperations()
        self.__public_cloud_name = self.__environment_variables_dict.get('PUBLIC_CLOUD_NAME', '')
        self.__es_index_cro = self.__environment_variables_dict.get('CRO_ES_INDEX', '')
        self.__account_id = IAMOperations().get_aws_account_id_name()
        self.__ce_payer_index = self.__environment_variables_dict.get('CE_PAYER_INDEX')

    def get_account_budget_from_payer_ce_report(self):
        """
        This method returns the account budget from the payer ce reports
        Check policy cost_explorer_payer_billings
        :return:
        """
        query = {
            "query": {
                "bool": {
                    "must": [
                        {"term": {"CloudName.keyword": self.__public_cloud_name}},
                        {"term": {"AccountId.keyword": self.__account_id}},
                        {"term": {"Month": str(datetime.utcnow().year)}},
                    ]
                }
            },
            "size": 1
        }
        response = self.__cost_over_usage.es_operations.fetch_data_by_es_query(query=query, es_index=self.__ce_payer_index, search_size=1, limit_to_size=True)
        if response:
            return response[0].get('_source').get(self.ALLOCATED_BUDGET)
        return 0

    def get_total_account_usage_cost(self):
        """
        This method returns the total account budget till date for this year
        :return:
        """
        current_date = datetime.utcnow().date()
        start_date = datetime(current_date.year, 1, 1).date()
        cost_explorer_operations = self.__cost_over_usage.get_cost_explorer_operations()
        response = cost_explorer_operations.get_cost_and_usage_from_aws(start_date=str(start_date), end_date=str(current_date+timedelta(days=1)), granularity='MONTHLY')
        total_cost = cost_explorer_operations.get_filter_data(ce_data=response['ResultsByTime'], group_by=False)
        return total_cost

    @typeguard.typechecked
    @logger_time_stamp
    def get_user_cost_data(self, group_by_tag_name: str, group_by_tag_value: str, requested_date: datetime = '', forecast: bool = False, duration: int = 0, extra_filter_key_values: dict = None):
        """
        This method fetch data from the es_reports
        :param extra_filter_key_values:
        :param group_by_tag_value:
        :param group_by_tag_name:
        :param duration:
        :param forecast:
        :param requested_date:
        :return:
        """
        extra_filter_matches = [{'Tags': {'Key': group_by_tag_name, 'Values': [group_by_tag_value]}}]
        if extra_filter_key_values:
            extra_filter_matches.extend([{'Tags': {'Key': filter_key, 'Values': [filter_value]}} for filter_key, filter_value in extra_filter_key_values.items()])
        start_date = requested_date.replace(minute=self.ZERO, hour=self.ZERO, second=self.ZERO, microsecond=self.ZERO)
        if forecast:
            end_date = start_date + timedelta(days=duration)
            response = self.__cost_over_usage.get_forecast_cost_data(start_date=start_date, end_date=end_date,
                                                                     extra_matches=extra_filter_matches, extra_operation=self.AND, tag_name=group_by_tag_name)
            return_key = 'Forecast'
        else:
            response = self.__cost_over_usage.get_monthly_user_es_cost_data(start_date=start_date,
                                                                            end_date=datetime.utcnow().replace(microsecond=self.ZERO) + timedelta(days=1),
                                                                            extra_matches=extra_filter_matches, extra_operation=self.AND, tag_name=group_by_tag_name)
            return_key = 'Cost'
        if response:
            return round(response[self.ZERO].get(return_key), self.DEFAULT_ROUND_DIGITS)
        return self.ZERO

    @typeguard.typechecked
    @logger_time_stamp
    def prepare_instance_data(self, instance_data: list, user: str, ticket_id: str, user_cost: float,
                              cost_estimation: float, ticket_opened_date: datetime):
        """
        This method returns es data to upload
        :param instance_data:
        :param user:
        :param ticket_id:
        :param user_cost:
        :param cost_estimation:
        :param ticket_opened_date:
        :return: dict data
        """
        return {
            'cloud_name': self.__public_cloud_name.upper(),
            'account_name': self.__account_name,
            'region_name': instance_data[self.ZERO].get('region_name'),
            'user': user,
            'user_cro': instance_data[self.ZERO].get('user_cro'),
            'actual_cost': user_cost,
            'ticket_id': ticket_id,
            'ticket_id_state': 'in-progress',
            'estimated_cost': cost_estimation,
            'total_instances': len(instance_data),
            'monitored_days': (datetime.utcnow().date() - ticket_opened_date.date()).days,
            'ticket_opened_date': ticket_opened_date.date(),
            'duration': int(instance_data[self.ZERO].get('duration')),
            'approved_manager': instance_data[self.ZERO].get('approved_manager'),
            'user_manager': instance_data[self.ZERO].get('manager'),
            'project': instance_data[self.ZERO].get('project'),
            'owner': instance_data[self.ZERO].get('owner'),
            'total_spots': len([instance for instance in instance_data if instance.get('instance_plan').lower() == 'spot']),
            'total_ondemand': len([instance for instance in instance_data if instance.get('instance_plan').lower() == 'ondemand']),
            self.ALLOCATED_BUDGET: self.get_account_budget_from_payer_ce_report(),
            'instances': [f"{instance.get('instance_name')}, {instance.get('instance_id')}, "
                          f"{instance.get('instance_plan')}, "
                          f"{instance.get('instance_type')}, "
                          f"{instance.get('instance_state')}, {instance.get('instance_running_days')}" for instance in instance_data],
            'instance_types': [instance.get('instance_type') for instance in instance_data]
        }

    @typeguard.typechecked
    @logger_time_stamp
    def __prepare_update_es_data(self, source: dict, instance_data: list, user_cost: float, cost_estimation: float):
        """
        This method update the values of jira id data
        :param source:
        :param instance_data:
        :param user_cost:
        :param cost_estimation:
        :return: dict data
        """
        for instance in instance_data:
            index = [idx for idx, es_instance in enumerate(source.get('instances', [])) if instance.get('instance_id') in es_instance]
            running_days = instance.get('instance_running_days')
            if index:
                source['instances'][index[self.ZERO]] = f"{instance.get('instance_name')}, {instance.get('instance_id')}, " \
                                                f"{instance.get('instance_plan')}, {instance.get('instance_type')}, " \
                                                f"{instance.get('instance_state')}, {running_days}"
            else:
                source.setdefault('instances', []).append(f"{instance.get('instance_name')}, {instance.get('instance_id')}, "
                                                          f"{instance.get('instance_plan')}, {instance.get('instance_type')}, "
                                                          f"{instance.get('instance_state')}, {running_days}")
                source.setdefault('instance_types', []).append(instance.get('instance_type'))
                if instance.get('instance_plan', '').lower() == 'spot':
                    source['total_spots'] = source.get('total_spots', 0) + 1
                else:
                    if instance.get('instance_plan', '').lower() == 'ondemand':
                        source['total_ondemand'] = source.get('total_ondemand', 0) + 1
        MonitorTickets().verify_es_instances_state(es_data=source)
        if datetime.strptime(source.get('timestamp'), "%Y-%m-%dT%H:%M:%S.%f").date() != datetime.now().date():
            source['monitored_days'] = (datetime.utcnow().date() - source.get('ticket_opened_date')).days
        source['total_instances'] = len(source.get('instances', self.ZERO))
        source['duration'] = int(instance_data[self.ZERO].get('duration'))
        source['estimated_cost'] = round(cost_estimation, self.DEFAULT_ROUND_DIGITS)
        source['actual_cost'] = user_cost
        if instance_data[self.ZERO].get('user_cro') and source.get('user_cro') != instance_data[self.ZERO].get('user_cro'):
            source['user_cro'] = instance_data[self.ZERO].get('user_cro')
        if instance_data[self.ZERO].get('user') and source.get('user') != instance_data[self.ZERO].get('user'):
            source['user'] = instance_data[self.ZERO].get('user')
        source['timestamp'] = datetime.utcnow()
        if source.get('ticket_id_state') != 'in-progress':
            source['ticket_id_state'] = 'in-progress'
            source['approved_manager'] = instance_data[self.ZERO].get('approved_manager')
            source['user_manager'] = instance_data[self.ZERO].get('manager'),
            source['user_manager'] = instance_data[self.ZERO].get('manager'),
            source[self.ALLOCATED_BUDGET] = self.get_account_budget_from_payer_ce_report()
        return source

    @typeguard.typechecked
    @logger_time_stamp
    def __upload_cro_report_to_es(self, monitor_data: dict):
        """
        This method uploads the data to elastic search index and return the data
        :param monitor_data:
        :return:
        """
        upload_data = {}
        for ticket_id, instance_data in monitor_data.items():
            ticket_id = ticket_id.split('-')[-1]
            user = instance_data[self.ZERO].get('user')
            user_project = instance_data[self.ZERO].get('project')
            issue_description = self.jira_operations.get_issue_description(ticket_id=ticket_id, state='ANY')
            ticket_opened_date = issue_description.get('TicketOpenedDate')
            group_by_tag_name = self.COST_EXPLORER_TAGS[self.TICKET_ID_KEY]
            user_cost = self.get_user_cost_data(group_by_tag_name=group_by_tag_name, group_by_tag_value=ticket_id,
                                                requested_date=ticket_opened_date,
                                                extra_filter_key_values={'Project': user_project})
            duration = int(instance_data[self.ZERO].get('duration', 0))
            user_forecast = self.get_user_cost_data(group_by_tag_name=group_by_tag_name, group_by_tag_value=ticket_id, requested_date=datetime.utcnow(), extra_filter_key_values={'Project': user_project}, forecast=True, duration=duration)
            cost_estimation = float(instance_data[self.ZERO].get('estimated_cost', self.ZERO))
            if self.__cost_over_usage.es_operations.verify_elastic_index_doc_id(index=self.__cost_over_usage.es_index_cro, doc_id=ticket_id):
                es_data = self.__cost_over_usage.es_operations.get_es_data_by_id(id=ticket_id, index=self.__cost_over_usage.es_index_cro)
                es_data['_source']['ticket_opened_date'] = ticket_opened_date.date()
                es_data['_source']['forecast'] = user_forecast
                es_data['_source']['user'] = user
                source = self.__prepare_update_es_data(source=es_data.get('_source'), instance_data=instance_data, cost_estimation=cost_estimation, user_cost=user_cost)
                self.__cost_over_usage.es_operations.update_elasticsearch_index(index=self.__es_index_cro, id=ticket_id, metadata=source)
                upload_data[ticket_id] = source
            else:
                if ticket_id not in upload_data:
                    source = self.prepare_instance_data(instance_data=instance_data, ticket_id=ticket_id, cost_estimation=cost_estimation, user=user,  user_cost=user_cost, ticket_opened_date=ticket_opened_date)
                    source['ticket_opened_date'] = ticket_opened_date.date()
                    source['forecast'] = user_forecast
                    source['user'] = user
                    if not source.get(self.ALLOCATED_BUDGET):
                        source[self.ALLOCATED_BUDGET] = self.get_account_budget_from_payer_ce_report()
                    self.__cost_over_usage.es_operations.upload_to_elasticsearch(index=self.__es_index_cro, data=source, id=ticket_id)
                    upload_data[ticket_id] = source
        return upload_data

    @logger_time_stamp
    def update_in_progress_ticket_cost(self):
        """
        This method updates the in-progress tickets costs
        :return:
        """
        query = {"query": {"bool": {"must": [
                        {"term": {"cloud_name.keyword": self.__public_cloud_name}},
                        {"term": {"account_name.keyword": self.__account_name.upper()}},
                        {"term": {"ticket_id_state.keyword": "in-progress"}}
                    ]
                }}}
        in_progress_es_tickets = self.__cost_over_usage.es_operations.fetch_data_by_es_query(query=query, es_index=self.__es_index_cro)
        total_account_cost = self.get_total_account_usage_cost()
        for in_progress_ticket in in_progress_es_tickets:
            source_data = in_progress_ticket.get('_source')
            ticket_id = source_data.get(self.TICKET_ID_KEY)
            if source_data.get('account_name').lower() in self.__account_name.lower():
                ticket_opened_date = datetime.strptime(source_data.get('ticket_opened_date'), "%Y-%m-%d")
                duration = int(source_data.get('duration', 0))
                user_project = source_data.get('project')
                group_by_tag_name = self.COST_EXPLORER_TAGS[self.TICKET_ID_KEY]
                user_cost = self.get_user_cost_data(group_by_tag_name=group_by_tag_name, group_by_tag_value=ticket_id, requested_date=ticket_opened_date)
                user_forecast = self.get_user_cost_data(group_by_tag_name=group_by_tag_name, group_by_tag_value=ticket_id, requested_date=datetime.utcnow(), forecast=True, duration=duration)
                update_data = {'actual_cost': user_cost, 'forecast': user_forecast, 'timestamp': datetime.utcnow(), f'TotalCurrentUsage-{datetime.utcnow().year}': total_account_cost}
                if not source_data.get(self.ALLOCATED_BUDGET):
                    update_data[self.ALLOCATED_BUDGET] = self.get_account_budget_from_payer_ce_report()
                self.__cost_over_usage.es_operations.update_elasticsearch_index(index=self.__es_index_cro, metadata=update_data, id=ticket_id)

    @typeguard.typechecked
    @logger_time_stamp
    def run(self, monitor_data: dict):
        """
        This method runs data collection methods
        :param monitor_data:
        :return:
        """
        handler.setLevel(logging.WARN)
        result = self.__upload_cro_report_to_es(monitor_data=monitor_data)
        handler.setLevel(logging.WARN)
        return result




import asyncio
import json
import os.path
from datetime import datetime

import typeguard

from cloud_governance.common.logger.logger_time_stamp import logger_time_stamp

from cloud_governance.common.jira.jira import Jira
from cloud_governance.main.environment_variables import environment_variables


class JiraOperations:
    """
    This Class is used for jira operations for cloud-governance
    """

    REFINEMENT = 'Refinement'
    IN_PROGRESS = 'In Progress'
    JIRA_TRANSITION_IDS = {
        'NEW': 51, 'REFINEMENT': 61, 'INPROGRESS': 31, 'CLOSED': 41, 'ANY': 0
    }
    FILE_EXTENSION = '.json'

    def __init__(self):
        self.__environment_variables_dict = environment_variables.environment_variables_dict
        self.__jira_url = self.__environment_variables_dict.get('JIRA_URL').strip()
        self.__jira_username = self.__environment_variables_dict.get('JIRA_USERNAME').strip()
        self.__jira_token = self.__environment_variables_dict.get('JIRA_TOKEN').strip()
        self.__jira_queue = self.__environment_variables_dict.get('JIRA_QUEUE').strip()
        self.__cache_temp_dir = self.__environment_variables_dict.get('TEMPORARY_DIRECTORY', '').strip()
        self.__loop = asyncio.new_event_loop()
        self.__jira_object = Jira(url=self.__jira_url, username=self.__jira_username, token=self.__jira_token, ticket_queue=self.__jira_queue, loop=self.__loop)

    @typeguard.typechecked
    @logger_time_stamp
    def move_issue_state(self, ticket_id: str, state: str):
        """
        This method move the issue state
        :param ticket_id:
        :param state:
        :return:
        """
        if '-' in ticket_id:
            ticket_id = ticket_id.split('-')[-1]
        state_id = self.JIRA_TRANSITION_IDS.get(state.upper())
        return self.__loop.run_until_complete(self.__jira_object.post_transition(ticket=ticket_id, transition=state_id))

    @typeguard.typechecked
    @logger_time_stamp
    def get_issue(self, ticket_id: str):
        """
        This method returns the issue data
        :param ticket_id:
        :return:
        """
        return self.__loop.run_until_complete(self.__jira_object.get_ticket(ticket=ticket_id))

    @typeguard.typechecked
    @logger_time_stamp
    def return_cache_ticket_description(self, ticket_id: str):
        """
        This method checks ticket id is already fetched from jira api
        :param ticket_id:
        :return:
        """
        with open(f'{self.__cache_temp_dir}/{ticket_id}{self.FILE_EXTENSION}') as cache_ticket_description:
            result_data = json.load(cache_ticket_description)
            result_data['TicketOpenedDate'] = datetime.strptime(result_data.get('TicketOpenedDate'), "%Y-%m-%d %H:%M:%S")
            return result_data

    @typeguard.typechecked
    @logger_time_stamp
    def cache_ticket_description(self, ticket_id: str, ticket_description: dict):
        """
        This method saves the ticket_id description
        :param ticket_description:
        :param ticket_id:
        :return:
        """
        if self.__cache_temp_dir:
            with open(f'{self.__cache_temp_dir}/{ticket_id}{self.FILE_EXTENSION}', 'w') as cache_ticket_description:
                json.dump(ticket_description, cache_ticket_description, default=str)

    @typeguard.typechecked
    @logger_time_stamp
    def get_issue_description(self, ticket_id: str, state: str = '', sub_task: bool = False):
        """
        This method return the ticket data description
        :param ticket_id:
        :param state:
        :param sub_task:
        :return:
        """
        if '-' in ticket_id:
            ticket_id = ticket_id.split('-')[-1]
        if os.path.exists(f'{self.__cache_temp_dir}/{ticket_id}{self.FILE_EXTENSION}'):
            return self.return_cache_ticket_description(ticket_id=ticket_id)
        else:
            issue_data = self.get_issue(ticket_id=ticket_id)
            if issue_data:
                if issue_data['fields']['status']['name'] == self.REFINEMENT or self.JIRA_TRANSITION_IDS.get(state, -1) == 0 or (state == 'INPROGRESS' and issue_data['fields']['status']['name'] == self.IN_PROGRESS) or sub_task:
                    description_list = issue_data['fields']['description'].split('\n')
                    description_dict = {}
                    for description in description_list:
                        if description:
                            values = description.strip().split(':', 1)
                            if len(values) == 2:
                                key, value = values
                                description_dict[key.strip().replace(' ', '')] = value.strip()
                    if 'Project' in description_dict:
                        if description_dict['Project'] == "Other":
                            description_dict['Project'] = description_dict.get('Explanationof"Other"primaryproduct')
                            if not description_dict['Project']:
                                description_dict['Project'] = description_dict.get('Otherproductsbeingtested')
                            if not description_dict['Project']:
                                description_dict['Project'] = description_dict.get('Explanationof"Other"secondaryproduct')
                    description_dict['TicketOpenedDate'] = datetime.strptime(issue_data.get('fields').get('created').split('.')[0], "%Y-%m-%dT%H:%M:%S")
                    description_dict['JiraStatus'] = issue_data['fields']['status']['name']
                    self.cache_ticket_description(ticket_id=ticket_id, ticket_description=description_dict)
                    return description_dict
        return {}

    @logger_time_stamp
    def get_all_issues_in_progress(self):
        """
        This method get all issues which are in progress
        :return:
        """
        issues = self.__loop.run_until_complete(self.__jira_object.search_tickets(query={'Status': "'IN PROGRESS'"})).get('issues')
        ticket_ids = {}
        for issue in issues:
            if '[Clouds]' in issue['fields']['summary']:
                ticket_id = issue.get('key')
                description = self.beautify_issue_description(issue['fields']['description'])
                description['TicketOpenedDate'] = datetime.strptime(issue.get('fields').get('created').split('.')[0], "%Y-%m-%dT%H:%M:%S")
                ticket_ids[ticket_id] = description.get('Region')
        return ticket_ids

    @logger_time_stamp
    def beautify_issue_description(self, description):
        """
        This method beautify the issue description
        """
        description = description.split("\n")
        description_data = {}
        for index, line in enumerate(description):
            if line:
                if ':' in line:
                    key, value = line.strip().split(':', 1)
                    description_data[key.strip().replace(' ', '')] = value.strip()
                else:
                    description_data[index] = line.strip()
        return description_data

    @typeguard.typechecked
    @logger_time_stamp
    def get_ticket_id_sub_tasks(self, ticket_id: str, closed: bool = False):
        """
        This method returns th list of subtasks
        :param ticket_id:
        :param closed:
        :return:
        """
        ticket_id = ticket_id.split("-")[-1]
        jira_data = self.get_issue(ticket_id=ticket_id)
        if jira_data:
            sub_tasks_ids = []
            sub_tasks = jira_data.get('fields').get('subtasks')
            for sub_task in sub_tasks:
                fields = sub_task.get('fields')
                if fields.get('status').get('name') != 'Closed' or closed:
                    sub_tasks_ids.append(sub_task.get('key'))
            return sub_tasks_ids
        return []

    @typeguard.typechecked
    @logger_time_stamp
    def get_issue_sub_tasks_cost_estimation(self, ticket_id: str):
        """
        This method get issue cost estimation
        :param ticket_id:
        :return:
        """
        sub_tasks = self.get_ticket_id_sub_tasks(ticket_id=ticket_id, closed=True)
        cost_estimation = 0
        for sub_task in sub_tasks:
            description = self.get_issue_description(ticket_id=sub_task, sub_task=True)
            cost_estimation += float(description.get('CostEstimation', 0))
        return cost_estimation

    @typeguard.typechecked
    @logger_time_stamp
    def get_all_issues(self, ticket_status: str):
        """
        This method get all issues which are in progress
        :param ticket_status:
        :return:
        """
        issues = self.__loop.run_until_complete(
            self.__jira_object.search_tickets(query={'Status': f"'{ticket_status}'"})).get('issues')
        ticket_ids = {}
        for issue in issues:
            if '[Clouds]' in issue['fields']['summary']:
                ticket_id = issue.get('key')
                description = self.beautify_issue_description(issue['fields']['description'])
                description['TicketOpenedDate'] = datetime.strptime(issue.get('fields').get('created').split('.')[0], "%Y-%m-%dT%H:%M:%S")
                ticket_ids[ticket_id] = description
        return ticket_ids



import json
import os
from ast import literal_eval
from copy import deepcopy

import boto3

from cloud_governance.common.ldap.ldap_search import LdapSearch
from cloud_governance.common.logger.init_logger import logger
from cloud_governance.common.logger.logger_time_stamp import logger_time_stamp
from cloud_governance.common.mails.mail_message import MailMessage
from cloud_governance.common.mails.postfix import Postfix
from cloud_governance.main.environment_variables import environment_variables


class ZombieClusterCommonMethods:
    DAYS_TO_TRIGGER_RESOURCE_MAIL = 4
    DAYS_TO_DELETE_RESOURCE = 7

    def __init__(self, region: str, force_delete: bool = False):
        self.__environment_variables_dict = environment_variables.environment_variables_dict
        self.region = region
        self.dry_run = self.__environment_variables_dict.get('dry_run', 'yes')
        self.policy = self.__environment_variables_dict.get('policy', '')
        self.ec2_client = boto3.client('ec2', region_name=region)
        self.ec2_resource = boto3.resource('ec2', region_name=region)
        self.elb_client = boto3.client('elb', region_name=region)
        self.elbv2_client = boto3.client('elbv2', region_name=region)
        self.iam_client = boto3.client('iam', region_name=region)
        self.s3_client = boto3.client('s3')
        self.s3_resource = boto3.resource('s3')
        self.__ldap_host_name = self.__environment_variables_dict.get('LDAP_HOST_NAME', '')
        self._special_user_mails = self.__environment_variables_dict.get('special_user_mails', '{}')
        self._account_admin = self.__environment_variables_dict.get('account_admin', '')
        self.__email_alert = self.__environment_variables_dict.get('EMAIL_ALERT') if self.__environment_variables_dict.get('EMAIL_ALERT') else False
        self._ldap = LdapSearch(ldap_host_name=self.__ldap_host_name)
        self._mail = Postfix()
        self._mail_description = MailMessage()
        self._force_delete = self.__environment_variables_dict.get('FORCE_DELETE') if self.__environment_variables_dict.get('FORCE_DELETE') else force_delete

    def _literal_eval(self, data: any):
        tags = {}
        if data:
            tags = literal_eval(data)
        return tags

    def get_tag_name_from_tags(self, tags: list, tag_name: str):
        """
        This method returns tag_name from resource_tags
        @param tags:
        @param tag_name:
        @return:
        """
        if tags:
            for tag in tags:
                if tag.get('Key') == tag_name:
                    return tag.get('Value')
        return ''

    def get_zombie_cluster_user_tag(self, zombies: dict, resources: list, resource_id: str, tag: str = 'Tags'):
        """
        This method returns the zombie_cluster tag and with user_tag i.e {cluster_tag: user_name}
        @param zombies:
        @param resources:
        @param resource_id:
        @return:
        """
        zombie_cluster_user_tag = {}
        for resource in resources:
            zombie_id = resource.get(resource_id)
            if zombie_id in zombies:
                zombie_cluster_user_tag[zombies[zombie_id]] = self.get_tag_name_from_tags(tags=resource.get(tag),
                                                                                          tag_name='User')
        return zombie_cluster_user_tag

    def _get_tags_of_zombie_resources(self, resources: list, resource_id_name: str, zombies: dict, aws_service: str,
                                      aws_tag: str = 'Tags'):
        """
        This method returns tags of the resource i.e {resource_id: tags}
        @param resources:
        @param tags:
        @return:
        """
        resources_tags = {}
        for resource in resources:
            aws_tags = []
            resource_id = resource.get(resource_id_name)
            if aws_service == 'elbv1':
                try:
                    aws_tags = self.elb_client.describe_tags(LoadBalancerNames=[resource_id]).get('TagDescriptions')
                    if len(aws_tags) > 0:
                        aws_tags = aws_tags[0].get(aws_tag)
                except:
                    return []
            elif aws_service == 'elbv2':
                try:
                    aws_tags = self.elbv2_client.describe_tags(ResourceArns=[resource_id]).get('TagDescriptions')
                    if len(aws_tags) > 0:
                        aws_tags = aws_tags[0].get(aws_tag)
                except:
                    return []
            elif aws_service == 'role' and resource_id in zombies:
                try:
                    role_data = self.iam_client.get_role(RoleName=resource_id)['Role']
                    if role_data.get(aws_tag):
                        aws_tags = role_data.get(aws_tag)
                except:
                    return []
            elif aws_service == 'user' and resource_id in zombies:
                try:
                    user_data = self.iam_client.get_user(UserName=resource_id)['User']
                    if user_data.get(aws_tag):
                        aws_tags = user_data.get(aws_tag)
                except Exception as err:
                    return []
            elif aws_service == 'bucket' and resource_id in zombies:
                try:
                    bucket_data = self.s3_client.get_bucket_tagging(Bucket=resource_id)
                    if bucket_data.get(aws_tag):
                        aws_tags = bucket_data.get(aws_tag)
                except Exception as err:
                    return []
            else:
                if resource.get(aws_tag) and resource_id in zombies:
                    aws_tags = resource.get(aws_tag)
            if aws_tags:
                old_tags = deepcopy(aws_tags)
                tags = aws_tags[:]
                cluster_delete_days = self.get_tag_name_from_tags(tags=tags, tag_name='ClusterDeleteDays')
                if cluster_delete_days:
                    if resource_id in zombies:
                        tags = self.update_resource_tags(tags=tags, tag_name='ClusterDeleteDays',
                                                         tag_value=str(int(cluster_delete_days) + 1))
                    else:
                        tags = self.update_resource_tags(tags=tags, tag_name='ClusterDeleteDays', tag_value=str(0))
                else:
                    if resource_id in zombies:
                        tags = self.update_resource_tags(tags=tags, tag_name='ClusterDeleteDays', tag_value=str(1))
                if old_tags != tags:
                    try:
                        if aws_service == 'ec2':
                            self.ec2_client.create_tags(Resources=[resource_id], Tags=tags)
                        elif aws_service == 'elbv1':
                            self.elb_client.add_tags(LoadBalancerNames=[resource_id], Tags=tags)
                        elif aws_service == 'elbv2':
                            self.elbv2_client.add_tags(ResourceArns=[resource_id], Tags=tags)
                        elif aws_service == 'role':
                            self.iam_client.tag_role(RoleName=resource_id, Tags=tags)
                        elif aws_service == 'user':
                            self.iam_client.tag_user(UserName=resource_id, Tags=tags)
                        elif aws_service == 'bucket':
                            self.s3_client.put_bucket_tagging(Bucket=resource_id, Tagging={'TagSet': tags})
                    except:
                        return []
                if resource_id in zombies:
                    resources_tags[resource_id] = tags
        return resources_tags

    def update_resource_tags(self, tags: list, tag_name: str, tag_value: str):
        """
        This method updates tags of the resource
        @param tags:
        @param tag_name:
        @param tag_value:
        @param resource_id:
        @return:
        """
        found = False
        if tags:
            for tag in tags:
                if tag.get('Key') == tag_name:
                    tag['Value'] = str(tag_value)
                    found = True
            if not found:
                tags.append({'Key': tag_name, 'Value': str(tag_value)})
            return tags
        return [{'Key': tag_name, 'Value': str(tag_value)}]

    def get_cluster_delete_days(self, tags: list) -> int:
        """
        This method returns the ClusterDeleteDays tag
        @param tags:
        @return:
        """
        cluster_delete_days = self.get_tag_name_from_tags(tags=tags, tag_name='ClusterDeleteDays')
        if not cluster_delete_days:
            cluster_delete_days = 1
        else:
            cluster_delete_days = int(cluster_delete_days) + 1
        return cluster_delete_days

    def trigger_mail(self, tags: list, resource_id: str, days: int, resources: list, message_type: str):
        """
        This method send triggering mail
        @param message_type:
        @param resources:
        @param days:
        @param tags:
        @param resource_id:
        @return:
        """
        try:
            special_user_mails = self._literal_eval(self._special_user_mails)
            user, resource_name = self.get_tag_name_from_tags(tags=tags, tag_name='User'), self.get_tag_name_from_tags(
                tags=tags, tag_name='Name')
            if not resource_name:
                resource_name = self.get_tag_name_from_tags(tags=tags, tag_name='cg-Name')
            to = user if user not in special_user_mails else special_user_mails[user]
            ldap_data = self._ldap.get_user_details(user_name=to)
            cc = [self._account_admin, f'{ldap_data.get("managerId")}@redhat.com']
            name = to
            if ldap_data:
                name = ldap_data.get('displayName')
            file_name = os.path.join('/tmp', f'{resource_name.replace("/", "-")}.json')
            with open(file_name, 'w') as file:
                json.dump(resources, file, indent=4)
            subject, body = self._mail_description.zombie_cluster_mail_message(name=name, days=days,
                                                                               notification_days=self.DAYS_TO_TRIGGER_RESOURCE_MAIL,
                                                                               resource_name=resource_name, delete_days=self.DAYS_TO_DELETE_RESOURCE)
            self._mail.send_email_postfix(to=to, content=body, subject=subject, cc=cc, resource_id=resource_id, filename=file_name, message_type=message_type)
        except Exception as err:
            logger.info(err)

    def collect_notify_cluster_data(self, resource_data: dict, cluster_left_out_days: dict, notify_data: dict,
                                    delete_data: dict, cluster_data: dict, func_name: str):
        """
        This method store the notify_data, delete_data of clusters
        @param func_name:
        @param cluster_data:
        @param resource_data:
        @param cluster_left_out_days:
        @param notify_data:
        @param delete_data:
        @return:
        """
        notify_tag_data = {}
        delete_tag_data = {}
        cluster_tags = set()
        for resource_id, cluster_tag in resource_data.items():
            if cluster_tag in cluster_left_out_days:
                cluster_tags.add(cluster_tag)
                zombie_days = int(self.get_tag_name_from_tags(tags=cluster_left_out_days[cluster_tag], tag_name='ClusterDeleteDays'))
                if zombie_days == self.DAYS_TO_TRIGGER_RESOURCE_MAIL:
                    notify_tag_data.setdefault(cluster_tag, []).append(resource_id)
                elif zombie_days >= self.DAYS_TO_DELETE_RESOURCE:
                    delete_tag_data.setdefault(cluster_tag, []).append(resource_id)
                if cluster_tag not in cluster_data:
                    cluster_data[cluster_tag] = cluster_left_out_days[cluster_tag]
        for cluster_tag in cluster_tags:
            if cluster_tag in cluster_left_out_days:
                if cluster_tag in notify_tag_data:
                    notify_data.setdefault(cluster_tag, []).append({func_name: notify_tag_data[cluster_tag]})
                if cluster_tag in delete_tag_data:
                    delete_data.setdefault(cluster_tag, []).append({func_name: delete_tag_data[cluster_tag]})
        return notify_data, delete_data, cluster_data

    def send_mails_to_cluster_user(self, notify_data: dict, delete_data: dict, cluster_data: dict):
        """
        This method send mail to the user to notify cluster status
        @param cluster_data:
        @param notify_data:
        @param delete_data:
        @return:
        """
        if self.__email_alert:
            for cluster_tag, resource_ids in notify_data.items():
                self.update_resource_tags(tags=cluster_data[cluster_tag], tag_name='Name', tag_value=cluster_tag)
                self.trigger_mail(tags=cluster_data[cluster_tag], resource_id=cluster_tag,
                                  days=self.DAYS_TO_TRIGGER_RESOURCE_MAIL,
                                  resources=resource_ids, message_type='notification')
            for cluster_tag, resource_ids in delete_data.items():
                self.update_resource_tags(tags=cluster_data[cluster_tag], tag_name='Name', tag_value=cluster_tag)
                self.trigger_mail(tags=cluster_data[cluster_tag], resource_id=cluster_tag,
                                  days=self.DAYS_TO_DELETE_RESOURCE, resources=resource_ids, message_type='delete')

    def _check_zombie_cluster_deleted_days(self, resources: dict, cluster_left_out_days: dict, zombie: str, cluster_tag: str):
        """
        This method check the cluster delete days and return the clusters
        @param resources:
        @param cluster_left_out_days:
        @return:
        """
        delete_cluster_resource = False
        if resources:
            cluster_delete_days = int(self.get_tag_name_from_tags(tags=resources[zombie], tag_name='ClusterDeleteDays'))
            if cluster_tag not in cluster_left_out_days:
                cluster_left_out_days[cluster_tag] = resources[zombie]
            if cluster_delete_days >= self.DAYS_TO_DELETE_RESOURCE:
                delete_cluster_resource = True
        return cluster_left_out_days, delete_cluster_resource




import logging
from datetime import datetime, timedelta

import typeguard

from cloud_governance.common.clouds.aws.cost_explorer.cost_explorer_operations import CostExplorerOperations
from cloud_governance.common.clouds.aws.ec2.ec2_operations import EC2Operations
from cloud_governance.common.elasticsearch.elasticsearch_operations import ElasticSearchOperations
from cloud_governance.common.ldap.ldap_search import LdapSearch
from cloud_governance.common.logger.init_logger import logger, handler
from cloud_governance.common.logger.logger_time_stamp import logger_time_stamp
from cloud_governance.common.mails.mail_message import MailMessage
from cloud_governance.common.mails.postfix import Postfix
from cloud_governance.main.environment_variables import environment_variables


class CostOverUsage:
    """
    This class will monitors the cost explorer reports and sends alert to the user if they exceed specified amount
    """

    DEFAULT_ROUND_DIGITS = 3
    SEND_ALERT_DAY = 3
    FORECAST_GRANULARITY = 'MONTHLY'
    FORECAST_COST_METRIC = 'UNBLENDED_COST'
    OVER_USAGE_THRESHOLD = 0.05
    CLOUD_GOVERNANCE_ES_MAIL_INDEX = 'cloud-governance-mail-messages'
    CRO_OVER_USAGE_ALERT = 'cro-over-usage-alert'
    TIMESTAMP_DATE_FORMAT = "%Y-%m-%dT%H:%M:%S.%f"

    def __init__(self):
        self.__environment_variables_dict = environment_variables.environment_variables_dict
        self.__aws_account = self.__environment_variables_dict.get('account', '').replace('OPENSHIFT-', '').strip()
        self.__postfix_mail = Postfix()
        self.__mail_message = MailMessage()
        self.__es_host = self.__environment_variables_dict.get('es_host', '')
        self.__es_port = self.__environment_variables_dict.get('es_port', '')
        self.__over_usage_amount = self.__environment_variables_dict.get('CRO_COST_OVER_USAGE', '')
        self.__es_ce_reports_index = self.__environment_variables_dict.get('USER_COST_INDEX', '')
        self.__ldap_search = LdapSearch(ldap_host_name=self.__environment_variables_dict.get('LDAP_HOST_NAME', ''))
        self.__cro_admins = self.__environment_variables_dict.get('CRO_DEFAULT_ADMINS', [])
        self.es_index_cro = self.__environment_variables_dict.get('CRO_ES_INDEX', '')
        self.__cro_duration_days = self.__environment_variables_dict.get('CRO_DURATION_DAYS')
        self.current_end_date = datetime.utcnow()
        self.current_start_date = self.current_end_date - timedelta(days=self.__cro_duration_days)
        self.__public_cloud_name = self.__environment_variables_dict.get('PUBLIC_CLOUD_NAME')
        self.__ce_operations = CostExplorerOperations()
        self.es_operations = ElasticSearchOperations(es_host=self.__es_host, es_port=self.__es_port)
        self.__over_usage_threshold = self.OVER_USAGE_THRESHOLD * self.__over_usage_amount
        self.__ec2_operations = EC2Operations()

    def get_cost_explorer_operations(self):
        return self.__ce_operations

    @typeguard.typechecked
    @logger_time_stamp
    def get_cost_based_on_tag(self, start_date: str, end_date: str, tag_name: str, extra_filters: any = None, extra_operation: str = 'And', granularity: str = None, forecast: bool = False):
        """
        This method gives the cost based on the tag_name
        :param forecast:
        :param granularity:
        :param extra_operation: default, And
        :param extra_filters:
        :param tag_name:
        :param start_date:
        :param end_date:
        :return:
        """
        # remove_savings_cost = {  # removed the savings plan usage from the user costs
        #     'Not': {
        #         'Dimensions': {
        #             'Key': 'RECORD_TYPE',
        #             'Values': ['SavingsPlanRecurringFee', 'SavingsPlanNegation', 'SavingsPlanCoveredUsage']
        #         }
        #     }
        # }
        Filters = {} #remove_savings_cost
        if extra_filters:
            if type(extra_filters) == list:
                if len(extra_filters) == 1:
                    Filters = extra_filters[0]
                else:
                    Filters = {
                        extra_operation: [
                            *extra_filters,
                            # remove_savings_cost
                        ]
                    }
            else:
                Filters = {
                    # extra_operation: [
                        extra_filters
                        # remove_savings_cost
                    # ]
                }
        if forecast:
            results_by_time = self.__ce_operations.get_cost_forecast(start_date=start_date, end_date=end_date, granularity=self.FORECAST_GRANULARITY, cost_metric=self.FORECAST_COST_METRIC, Filter=Filters)['Total']
            response = [{'Forecast': round(float(results_by_time.get('Amount')), self.DEFAULT_ROUND_DIGITS)}]
        else:
            results_by_time = self.__ce_operations.get_cost_by_tags(start_date=start_date, end_date=end_date, tag=tag_name, Filter=Filters, granularity=granularity)['ResultsByTime']
            response = self.__ce_operations.get_filter_data(ce_data=results_by_time, tag_name=tag_name)
        return response

    @typeguard.typechecked
    @logger_time_stamp
    def __get_start_end_dates(self, start_date: datetime = None, end_date: datetime = None):
        """
        This method returns the start_date and end_date
        :param start_date:
        :param end_date:
        :return:
        """
        if not start_date and not end_date:
            end_date = self.current_end_date.date()
            start_date = self.current_start_date.date()
        elif not start_date:
            start_date = self.current_start_date.date()
            end_date = end_date.date()
        else:
            if not end_date:
                end_date = self.current_end_date.date()
            else:
                end_date = end_date.date()
            start_date = start_date.date()
        return start_date, end_date

    @typeguard.typechecked
    @logger_time_stamp
    def get_monthly_user_es_cost_data(self, tag_name: str = 'User', start_date: datetime = None, end_date: datetime = None, extra_matches: any = None, granularity: str = 'MONTHLY', extra_operation: str = 'And'):
        """
        This method gets the user cost from the es-data
        :param tag_name: by default User
        :param start_date:
        :param end_date:
        :param extra_matches:
        :param granularity: by default MONTHLY
        :param extra_operation:
        :return:
        """
        start_date, end_date = self.__get_start_end_dates(start_date=start_date, end_date=end_date)
        return self.get_cost_based_on_tag(start_date=str(start_date), end_date=str(end_date), tag_name=tag_name, granularity=granularity, extra_filters=extra_matches, extra_operation=extra_operation)

    def get_forecast_cost_data(self, tag_name: str = 'User', start_date: datetime = None, end_date: datetime = None, extra_matches: any = None, granularity: str = 'MONTHLY', extra_operation: str = 'And'):
        """
        This method returns the forecast based on inputs
        :param tag_name: by default User
        :param start_date:
        :param end_date:
        :param extra_matches:
        :param granularity: by default MONTHLY
        :param extra_operation:
        :return:
        """
        start_date, end_date = self.__get_start_end_dates(start_date=start_date, end_date=end_date)
        return self.get_cost_based_on_tag(start_date=str(start_date), end_date=str(end_date), tag_name=tag_name, granularity=granularity, extra_filters=extra_matches, extra_operation=extra_operation, forecast=True)

    def get_user_active_ticket_costs(self, user_name: str):
        """
        This method returns a boolean indicating whether the user should open the ticket or not
        :param user_name:
        :return:
        """
        query = {  # check user opened the ticket in elastic_search
            "query": {
                "bool": {
                    "must": [{"term": {"user_cro.keyword": user_name}},
                             {"terms": {"ticket_id_state.keyword": ['new', 'manager-approved', 'in-progress']}},
                             {"term": {"account_name.keyword": self.__aws_account.upper()}}
                             ],
                    "filter": {
                        "range": {
                            "timestamp": {
                                "format": "yyyy-MM-dd",
                                "lte": str(self.current_end_date.date()),
                                "gte": str(self.current_start_date.date()),
                            }
                        }
                    }
                }
            }
        }
        user_active_tickets = self.es_operations.fetch_data_by_es_query(es_index=self.es_index_cro, query=query)
        if not user_active_tickets:
            return None
        else:
            total_active_ticket_cost = 0
            for cro_data in user_active_tickets:
                opened_ticket_cost = float(cro_data.get('_source').get('estimated_cost'))
                total_active_ticket_cost += opened_ticket_cost
            return total_active_ticket_cost

    @logger_time_stamp
    def get_cost_over_usage_users(self):
        """
        This method returns the cost over usage users which are not opened ticket
        :return:
        """
        over_usage_users = []
        current_month_users = self.get_monthly_user_es_cost_data()
        for user in current_month_users:
            user_name = str(user.get('User'))
            user_cost = round(user.get('Cost'), self.DEFAULT_ROUND_DIGITS)
            if user_cost >= (self.__over_usage_amount - self.__over_usage_threshold):
                user_active_tickets_cost = self.get_user_active_ticket_costs(user_name=user_name)
                if not user_active_tickets_cost:
                    over_usage_users.append(user)
                else:
                    user_cost_without_active_ticket = user_cost - user_active_tickets_cost
                    if user_cost_without_active_ticket > self.__over_usage_amount:
                        user['Cost'] = user_cost_without_active_ticket
                        over_usage_users.append(user)
        return over_usage_users

    @logger_time_stamp
    def send_alerts_to_over_usage_users(self):
        """
        This method send alerts to cost over usage users
        Send alert to users every 3rd day, if not open ticket
        :return:
        """
        users_list = self.get_cost_over_usage_users()
        alerted_users = []
        for row in users_list:
            user, cost, project = row.get('User'), row.get('Cost'), row.get('Project', '')
            # send_alert, alert_number = self.get_last_mail_alert_status(user=str(user))
            # if send_alert:
            alerted_users.append(user)
            cc = [*self.__cro_admins]
            user_details = self.__ldap_search.get_user_details(user_name=user)
            if user_details:
                if self.__ec2_operations.verify_active_instances(tag_name='User', tag_value=str(user)):
                    name = f'{user_details.get("FullName")}'
                    cc.append(user_details.get('managerId'))
                    subject, body = self.__mail_message.cro_cost_over_usage(CloudName=self.__public_cloud_name,
                                                                            OverUsageCost=self.__over_usage_amount,
                                                                            FullName=name, Cost=cost, Project=project, to=user)
                    es_data = {'Alert': 1, 'MissingUserTicketCost': cost}
                    handler.setLevel(logging.WARN)
                    self.__postfix_mail.send_email_postfix(to=user, cc=[], content=body, subject=subject, mime_type='html', es_data=es_data, message_type=self.CRO_OVER_USAGE_ALERT)
                    handler.setLevel(logging.INFO)
        return alerted_users

    def get_last_mail_alert_status(self, user: str):
        """
        This method return the last mail alert.
        :param user:
        :return:
        """
        query = {
            "query": {
                "bool": {
                    "must": [
                        {"term": {"To.keyword": user}},
                        {"term": {"MessageType.keyword": self.CRO_OVER_USAGE_ALERT}},
                    ]
                }
            },
            "size": 1,
            "sort": {"timestamp": "desc"}
        }
        response = self.es_operations.fetch_data_by_es_query(query=query, es_index=self.CLOUD_GOVERNANCE_ES_MAIL_INDEX, search_size=1, limit_to_size=True)
        if response:
            last_alert = response[0]
            last_send_date = last_alert.get('_source').get('timestamp')
            alert_number = last_alert.get('_source').get('Alert', 0)
            current_date = datetime.utcnow().date()
            last_send_date = datetime.strptime(last_send_date, self.TIMESTAMP_DATE_FORMAT).date()
            days = (current_date - last_send_date).days
            if days % self.SEND_ALERT_DAY == 0 and last_send_date != current_date:
                return True, alert_number
            logger.warning(f"Already sent mail on {last_send_date} to {user}")
            return False, alert_number
        return True, 0

    @logger_time_stamp
    def run(self):
        """
        This method runs the cost over usage methods
        :return:
        """
        return self.send_alerts_to_over_usage_users()



import calendar
import datetime
import time

import pytz
from azure.mgmt.costmanagement.models import QueryDataset, QueryAggregation, QueryTimePeriod, QueryGrouping

from cloud_governance.common.clouds.azure.subscriptions.azure_operations import AzureOperations
from cloud_governance.common.logger.init_logger import logger
from cloud_governance.common.logger.logger_time_stamp import logger_time_stamp


class CostManagementOperations:
    """This class for fetching the azure usage and forecast reports"""

    def __init__(self):
        self.azure_operations = AzureOperations()

    @logger_time_stamp
    def get_usage(self, scope: str, start_date: datetime = '', end_date: datetime = '', granularity: str = 'Monthly', **kwargs):
        """
        This method get the current usage based on month
        @return:
        """
        try:
            if not start_date and not end_date:
                end_date = datetime.datetime.now(pytz.UTC)
                start_date = (end_date - datetime.timedelta(days=30)).replace(day=1)
            response = self.azure_operations.cost_mgmt_client.query.usage(scope=scope, parameters={
                'type': 'Usage', 'timeframe': 'Custom', 'time_period': QueryTimePeriod(from_property=start_date, to=end_date),
                'dataset': QueryDataset(granularity=granularity, aggregation={
                    "totalCost": QueryAggregation(name="Cost", function="Sum")}, **kwargs
                )
            })
            return response.as_dict()
        except Exception as e:
            print(e)
        return []

    @logger_time_stamp
    def get_forecast(self, scope: str, start_date: datetime = '', end_date: datetime = '', granularity: str = 'Monthly', **kwargs):
        """
        This method gets the forecast of next couple of months
        @param start_date:
        @param end_date:
        @param granularity:
        @param scope:
        @return:
        """
        try:
            if not start_date and not end_date:
                current = datetime.datetime.now(pytz.UTC)
                start_date = (current - datetime.timedelta(60)).replace(day=1)
                end_date = (current + datetime.timedelta(365))
                month_end = calendar.monthrange(end_date.year, end_date.month)[1]
                end_date = end_date.replace(day=month_end)
            logger.info(f'StartDate: {start_date}, EndDate: {end_date}')
            response = self.azure_operations.cost_mgmt_client.forecast.usage(scope=scope, parameters={
                'type': 'ActualCost', 'timeframe': 'Custom',
                'time_period': QueryTimePeriod(from_property=start_date, to=end_date),
                'dataset': QueryDataset(granularity=granularity, aggregation={
                    "totalCost": QueryAggregation(name="Cost", function="Sum"),
                }, **kwargs), 'include_actual_cost': True, 'include_fresh_partial_cost': False
            }).as_dict()
            result = {'columns': response.get('columns'), 'rows': []}
            row_data = {}
            for data in response.get('rows'):
                data_date = data[1]
                if data_date in row_data:
                    if row_data[data_date][2] == 'Actual' and data[2] == 'Forecast':
                        row_data[data_date][2] = data[2]
                    row_data[data_date][0] += data[0]
                else:
                    row_data[data_date] = data
            result['rows'] = list(row_data.values())
            return result
        except Exception as err:
            logger.info(f'Error raised when fetching the forecasting results {err}')
            return []



from datetime import datetime

from cloud_governance.common.clouds.aws.ec2.ec2_operations import EC2Operations
from cloud_governance.common.logger.logger_time_stamp import logger_time_stamp
from cloud_governance.main.environment_variables import environment_variables


class MonitorCROInstances:
    """
    This class monitor cro instances
    """

    def __init__(self, region_name: str = ''):
        self.__environment_variables_dict = environment_variables.environment_variables_dict
        self.__region_name = region_name if region_name else self.__environment_variables_dict.get('AWS_DEFAULT_REGION')
        self.__ec2_operations = EC2Operations(region=self.__region_name)
        self.__cro_resource_tag_name = self.__environment_variables_dict.get('CRO_RESOURCE_TAG_NAME')

    @logger_time_stamp
    def __monitor_instances(self):
        """
        This method monitor instances and returns the data
        :return:
        """
        monitored_ticket_ids = {}
        filters = {'Filters': [{'Name': 'tag-key', 'Values': ['Duration']}]}
        instances = self.__ec2_operations.get_ec2_instance_list(**filters)
        for resource in instances:
            tags = resource.get('Tags')
            create_time = self.__ec2_operations.get_attached_time(volume_list=resource.get('BlockDeviceMappings'))
            if not create_time:
                create_time = resource.get('LaunchTime')
            ticket_id = self.__ec2_operations.get_tag_value_from_tags(tag_name=self.__cro_resource_tag_name, tags=tags)
            running_days = (datetime.now().date() - create_time.date()).days
            monitored_ticket_ids.setdefault(ticket_id, []).append({
                'region_name': self.__region_name,
                'ticket_id': ticket_id,
                'instance_id': resource.get('InstanceId'),
                'instance_create_time': create_time,
                'instance_state': resource.get('State')['Name'],
                'instance_type': resource.get('InstanceType'),
                'instance_running_days': running_days,
                'instance_plan': resource.get('InstanceLifecycle', 'ondemand'),
                'user_cro': self.__ec2_operations.get_tag_value_from_tags(tag_name='UserCRO', tags=tags),
                'user': self.__ec2_operations.get_tag_value_from_tags(tag_name='User', tags=tags),
                'manager': self.__ec2_operations.get_tag_value_from_tags(tag_name='Manager', tags=tags),
                'approved_manager': self.__ec2_operations.get_tag_value_from_tags(tag_name='ApprovedManager', tags=tags),
                'owner': self.__ec2_operations.get_tag_value_from_tags(tag_name='Owner', tags=tags),
                'project': self.__ec2_operations.get_tag_value_from_tags(tag_name='Project', tags=tags),
                'instance_name': self.__ec2_operations.get_tag_value_from_tags(tag_name='Name', tags=tags),
                'email': self.__ec2_operations.get_tag_value_from_tags(tag_name='Email', tags=tags),
                'duration': self.__ec2_operations.get_tag_value_from_tags(tag_name='Duration', tags=tags, cast_type='int'),
                'estimated_cost': self.__ec2_operations.get_tag_value_from_tags(tag_name='EstimatedCost', tags=tags, cast_type='float')
            })
        return monitored_ticket_ids

    @logger_time_stamp
    def run(self):
        """
        This method run the monitoring methods
        :return:
        """
        return self.__monitor_instances()



import importlib
import inspect

from cloud_governance.common.logger.init_logger import logger
from cloud_governance.policy.policy_operations.aws.zombie_non_cluster.run_zombie_non_cluster_policies import NonClusterZombiePolicy


class ZombieNonClusterPolicies(NonClusterZombiePolicy):

    def __init__(self):
        super().__init__()

    def run(self):
        """
        This method run the zombie non-cluster policies class
        @return:
        """
        logger.info(f'account={self._account}, region={self._region}, policy={self._policy}, dry_run={self._dry_run}')
        zombie_non_cluster_policy_module = importlib.import_module(f'cloud_governance.policy.aws.{self._policy}')

        for cls in inspect.getmembers(zombie_non_cluster_policy_module, inspect.isclass):
            if self._policy.replace('_', '') == cls[0].lower():
                response = cls[1]().run()
                if isinstance(response, str):
                    logger.info(f'key: {cls[0]}, Response: {response}')
                else:
                    logger.info(f'key: {cls[0]}, count: {len(response)}, {response}')
                    policy_result = response
                    if self._policy_output:
                        # if self._policy not in ('ec2_idle', 'ebs_in_use', 'ec2_run', 's3_inactive', 'zombie_snapshots', 'nat_gateway_unused'):
                        #     beautify_data = self._beautify_upload_data(upload_resource_data=response)
                        #     policy_result = {'count': len(beautify_data), self._policy: beautify_data}
                        logger.info(policy_result)
                        self._s3operations.save_results_to_s3(policy=self._policy.replace('_', '-'), policy_output=self._policy_output, policy_result=policy_result)
